# -*- coding: utf-8 -*-
"""
Created on Tue Jul 31 15:04:19 2018

@author: xavier.mouy
"""

import numpy as np
import pandas as pd
import mpl_toolkits.mplot3d
import matplotlib.pyplot as plt
import time
import os
from ecosound.core.audiotools import upsample
import scipy.signal
from numba import njit


def defineSphereSurfaceGrid(npoints, radius, origin=[0, 0, 0]):
    # Using the golden spiral method
    # ------------------
    # inputs:
    #   npoints =>  nb of points on the sphere - integer
    #   radius  => radius of the sphere - float
    #   origin  => origin of teh sphere in catesian coordinates- 3 element list
    # ------------------
    # sampling in spherical coordinates
    indices = np.arange(0, npoints, dtype=float) + 0.5
    phi = np.arccos(1 - 2*indices/npoints)
    theta = np.pi * (1 + 5**0.5) * indices
    # convert to cartesian coordinates
    Sx, Sy, Sz = radius*np.cos(theta) * np.sin(phi), radius*np.sin(theta) * np.sin(phi), radius*np.cos(phi)
    # Adjust origin
    Sx = Sx + origin[0]
    Sy = Sy + origin[1]
    Sz = Sz + origin[2]
    # package in a datafrane
    S = pd.DataFrame({'x': Sx, 'y': Sy, 'z': Sz})
    return S


def defineSphereVolumeGrid(spacing, radius, origin=[0, 0, 0]):
    # ------------------
    # inputs:
    #   spacing =>  distance in meters separatying each receiver - float
    #   radius  => radius of the sphere - float
    #   origin  => origin of the sphere in catesian coordinates- 3 element list
    # ------------------
    # Cube of points (Cartesian coordinates)
    vec = np.arange(-radius, radius+spacing, spacing)
    X, Y, Z = np.meshgrid(vec, vec, vec, indexing='ij')
    Sx = np.reshape(X, X.shape[0]*X.shape[1]*X.shape[2])
    Sy = np.reshape(Y, Y.shape[0]*Y.shape[1]*Y.shape[2])
    Sz = np.reshape(Z, Z.shape[0]*Z.shape[1]*Z.shape[2])
    # Convert to spherical coordinates and remove points with r < radius
    Sr = np.sqrt(Sx**2 + Sy**2 + Sz**2)
    Sr_sphere = Sr <= radius
    Sx_sphere = Sx[Sr_sphere]
    Sy_sphere = Sy[Sr_sphere]
    Sz_sphere = Sz[Sr_sphere]
    # Adjust origin
    Sx_sphere = Sx_sphere + origin[0]
    Sy_sphere = Sy_sphere + origin[1]
    Sz_sphere = Sz_sphere + origin[2]
    # package in a datafrane
    S = pd.DataFrame({'x': Sx_sphere, 'y': Sy_sphere, 'z': Sz_sphere})
    return S


def defineCubeVolumeGrid(spacing, radius, origin=[0, 0, 0]):
    # ------------------
    # inputs:
    #   spacing =>  distance in meters separatying each receiver - float
    #   radius  => radius of the sphere - float
    #   origin  => origin of the sphere in catesian coordinates- 3 element list
    # ------------------
    # Cube of points (Cartesian coordinates)
    vec = np.arange(-radius, radius+spacing, spacing)
    X, Y, Z = np.meshgrid(vec, vec, vec, indexing='ij')
    Sx = np.reshape(X, X.shape[0]*X.shape[1]*X.shape[2])
    Sy = np.reshape(Y, Y.shape[0]*Y.shape[1]*Y.shape[2])
    Sz = np.reshape(Z, Z.shape[0]*Z.shape[1]*Z.shape[2])
    # Adjust origin
    Sx = Sx + origin[0]
    Sy = Sy + origin[1]
    Sz = Sz + origin[2]
    # package in a datafrane
    S = pd.DataFrame({'x': Sx, 'y': Sy, 'z': Sz})
    return S


def defineReceiverPairs (n_receivers, ref_receiver=0):
    Rpairs = []
    for i in range(n_receivers):
        if i != ref_receiver:
            pair = [ref_receiver, i]
            Rpairs.append(pair)
    return Rpairs

def defineJacobian(R, S, V, Rpairs):
    unknowns = ['x','y','z']    # unknowns: 3D coordinates of sound source
    N = R.shape[0] - 1          # nb of measurements (TDOAs)
    M = len(unknowns)           # number of model parameters (unknowns)
    nsources = S.shape[0]       # number of sources
    J = [None] * nsources       # initiaization
    # for each source location
    for idx in range(nsources):
        s = S.iloc[idx]
        j = np.full([N, M], np.nan)  # initialization of Jacobian for that source location
        for i in range(N):
            p1 = Rpairs[i][0]        # receiver #1 ID
            p2 = Rpairs[i][1]        # receiver #2 ID
            for kk, unknown in enumerate(unknowns):
                Term1 = (1/V)*0.5*((((s.x-R.x[p1])**2)+((s.y-R.y[p1])**2)+((s.z-R.z[p1])**2))**(-0.5))*2*(s[unknown]-R[unknown][p1])
                Term2 = (1/V)*0.5*((((s.x-R.x[p2])**2)+((s.y-R.y[p2])**2)+((s.z-R.z[p2])**2))**(-0.5))*2*(s[unknown]-R[unknown][p2])
                j[i][kk] = Term2 - Term1
        J[idx] = j  # stacks jacobians for each source
    if nsources == 1:
        J = J[0]
    return J

def predict_tdoa(m, V, hydrophones_coords, hydrophone_pairs):
    """ Create data from the forward problem.
        Generate TDOAs based on location of hydrophones and source.
    """
    N = len(hydrophone_pairs)
    dt = np.full([N,1], np.nan)
    for idx, hydrophone_pair in enumerate(hydrophone_pairs):
        p1=hydrophone_pair[0]
        p2=hydrophone_pair[1]
        t1=((1/V)*np.sqrt( ((m.x-hydrophones_coords.x[p1])**2)+((m.y-hydrophones_coords.y[p1])**2)+((m.z-hydrophones_coords.z[p1])**2)) )
        t2=((1/V)*np.sqrt(((m.x-hydrophones_coords.x[p2])**2)+((m.y-hydrophones_coords.y[p2])**2)+((m.z-hydrophones_coords.z[p2])**2)))
        dt[idx] = np.array(t2-t1) # noiseless data
    return dt


def getUncertainties(J, NoiseVariance):
    nsources = len(J)
    errLoc_X = [None] * nsources
    errLoc_Y = [None] * nsources
    errLoc_Z = [None] * nsources
    errLoc_RMS = [None] * nsources
    for i in range(nsources):
        Cm = NoiseVariance * np.linalg.inv(np.dot(np.transpose(J[i]), J[i]))  # covariance matrix of the model
        errLoc_X[i], errLoc_Y[i], errLoc_Z[i] = np.sqrt(np.diag(Cm))  # uncertainty (std) along each axis
        errLoc_RMS[i] = np.sqrt(errLoc_X[i]**2 + errLoc_Y[i]**2 + errLoc_Z[i]**2)  # overall uncertainty (RMS)
    Uncertainty = pd.DataFrame({'x': errLoc_X, 'y': errLoc_Y, 'z': errLoc_Z, 'rms': errLoc_RMS})
    return Uncertainty


def plotArrayUncertainties(R, S, Uncertainties):
    fig1 = plt.figure()
    ax1 = fig1.add_subplot(111, projection='3d')
    # Sources
    points = ax1.scatter(S['x'], S['y'], S['z'], s=10, c=Uncertainties.rms.values, cmap='Spectral')
    cbar = plt.colorbar(points)
    cbar.ax.set_ylabel('Uncertainty (m)')
    # Receivers
    ax1.scatter(R['x'], R['y'], R['z'], s=30, c='black')
    # Axes labels
    ax1.set_xlabel('X (m)', labelpad=10)
    ax1.set_ylabel('Y (m)', labelpad=10)
    ax1.set_zlabel('Z (m)', labelpad=10)
    plt.show()


def getCost(R, S, Rpairs, V, NoiseVariance):
    # Get list of Jacobian matrice for each source
    J = defineJacobian(R, S, V, Rpairs)
    # Calculates localization uncertainty for each source
    Uncertainties = getUncertainties(J, NoiseVariance)
    # Get max uncertainty
    #E = max(Uncertainties.rms)
    E = np.mean(Uncertainties.rms)
    #E = np.median(Uncertainties.rms)
    return E


def getReceiverBoundsWidth(ReceiverBounds):
    ReceiverBoundsWidth = ReceiverBounds.applymap(lambda x: max(x)-min(x))
    return ReceiverBoundsWidth


def initializeReceivers(nReceivers, ReceiverBounds):
    ReceiverBoundsWidth = getReceiverBoundsWidth(ReceiverBounds)
    R1 = [None] * ReceiverBounds.shape[1]
    R1[0] = [None] * nReceivers  # x
    R1[1] = [None] * nReceivers  # y
    R1[2] = [None] * nReceivers  # z
    for r in range(nReceivers):  # random location for each receiver and axis
        for dim in range(ReceiverBounds.shape[1]):
            R1[dim][r] = np.random.rand(1)[0]*ReceiverBoundsWidth.iloc[r][dim] + min(ReceiverBounds.iloc[r][dim])
    Receivers = pd.DataFrame({'x': R1[0], 'y': R1[1], 'z': R1[2]})
    return Receivers


def getParamsLinearMapping(R):
    Rindices = [None] * R.shape[0] * R.shape[1]
    midx = 0
    for ridx in range(R.shape[0]):
        for dimidx in range(R.shape[1]):
            Rindices[midx] = [ridx, dimidx]
            midx += 1
    return Rindices


def perturbReceivers(R, PerturbParamIdx, MappedParamsIdx, ReceiverBounds, ReceiverBoundsWidth, PerturbSTD, T0, T):
    # goes back to first parameters if reached the end of the list of parameters
    if PerturbParamIdx > len(MappedParamsIdx)-1:
        PerturbParamIdx = 0
    # Identifies from MappedParamsIdx which Receiver and Dimension to perturb
    rid = MappedParamsIdx[PerturbParamIdx][0]    # Receiver ID
    dimid = MappedParamsIdx[PerturbParamIdx][1]  # Dimension ID
    # Add perturbation to parameter
    perturb = ((PerturbSTD*ReceiverBoundsWidth.iloc[rid][dimid])*np.random.normal(loc=0))  # Gaussian distributed perturbation
    # nu = np.random.random()
    # gamma = PerturbSTD*(T/T0)*np.tan(np.pi*(nu-0.5))
    # bandwidth = ReceiverBoundsWidth.iloc[rid][dimid]
    # perturb = gamma*bandwidth
    newparam = R.iloc[rid][dimid] + perturb
    # print(perturb)
    # Checks that perturbed parameter lies within the bounds
    isinbound = (newparam >= min(ReceiverBounds.iloc[rid][dimid])) & (newparam <= max(ReceiverBounds.iloc[rid][dimid]))
    # updates receiver parameter (only if new paramater fall within parameter bounds)
    R_prime = pd.DataFrame.copy(R)
    if isinbound == True:
        R_prime.iloc[rid][dimid] = newparam
    return R_prime, isinbound, PerturbParamIdx


def optimizeArray(ReceiverBounds, nReceivers, AnnealingSchedule, S, Rpairs, V, NoiseVariance):
    # start clock
    start = time.time()

    # Defines width of parameters bounds
    ReceiverBoundsWidth = getReceiverBoundsWidth(ReceiverBounds)

    # initialization of variables
    Cost = pd.DataFrame({'T': [], 'cost': []})
    acceptRateChanges = pd.DataFrame({'T': [], 'acceptRate': []})
    acceptRate = 1
    PerturbParamIdx = -1
    Tidx = 0  # temperature step index
    LoopStopFlag = 0
    Rchanges = []
    while LoopStopFlag == 0:  # Temperature loop. Keeps iterating until acceptance rate is too low
        # for nnn in range(1):
        # First iteration
        if (Tidx == 0):
            R = initializeReceivers(nReceivers, ReceiverBounds)  # random initialization of receivers locations (whithin the bounds)
            E_m = getCost(R, S, Rpairs, V, NoiseVariance)        # Calculates max RMS uncertainty
            T = AnnealingSchedule['Start']                       # initial temperature
            tmp1 = pd.DataFrame({'T': [T], 'cost': [E_m]})
            Cost = pd.DataFrame.append(Cost, tmp1, ignore_index=True)
            Rchanges = R.as_matrix()                                         # Keeps track of model paraneters at each iteration
            MappedParamsIdx = getParamsLinearMapping(R)          # linear list of each elements to optimize (for the perturnation phase)
            #plotArrayUncertainties(R, S, Uncertainties)

        # Checks that starting temperature is not set to low
        if (Tidx == 1):
            if (acceptRate < AnnealingSchedule['StartAcceptanceRate']):
                raise ValueError(['The acceptance rate during the melting phase is too low.Please adjust starting temperature (' + str(acceptRate) + ')'])
            else:
                print('Melting temperature valid (' + str(acceptRate) + ').')

        # Perturb paraneters
        nAccepted = 0  # keeps track of accepted perturbations
        for j in range(AnnealingSchedule['nPerturb']):  # perturbations of parameters for temperature T
            # Perturb one receiver parameter
            PerturbParamIdx += 1  # increment to next parameter in line
            R_prime, isinbound, PerturbParamIdx = perturbReceivers(R, PerturbParamIdx, MappedParamsIdx, ReceiverBounds, ReceiverBoundsWidth, AnnealingSchedule['PerturbSTD'], AnnealingSchedule['Start'], T)
            # Acceptance tests
            acceptedFlag = []
            if isinbound:
                E_mprime = getCost(R_prime, S, Rpairs, V, NoiseVariance)       # Calculates max RMS uncertainty
                deltaE = E_mprime-E_m
                if (deltaE <= 0):   # accept change
                    acceptedFlag = True
                elif (deltaE > 0):
                    psy = np.random.uniform()
                    P = np.exp(-deltaE/T)
                    if (psy <= P):  # accept change
                        acceptedFlag = True
                    else:           # reject change
                        acceptedFlag = False
            else:
                acceptedFlag = False
            # Updates parmaters for accepted changes
            if acceptedFlag:  # accepted change
                nAccepted += 1
                # update paramters
                R = pd.DataFrame.copy(R_prime)
                E_m = E_mprime
                del R_prime, E_mprime   # delete variables
                # saves cost and model parameters
                #tmp1 = pd.DataFrame({'T': [T], 'cost': [E_m]})
                #Cost = pd.DataFrame.append(Cost, tmp1, ignore_index=True)
                #Rchanges = np.dstack((Rchanges, R.as_matrix()))
            elif acceptedFlag is not False:  # sanity check
                raise ValueError(['Perturbation did go throught the acceptance test. Check the code for errors!'])
            # saves cost and model parameters
            tmp1 = pd.DataFrame({'T': [T], 'cost': [E_m]})
            Cost = pd.DataFrame.append(Cost, tmp1, ignore_index=True)
            Rchanges = np.dstack((Rchanges, R.as_matrix()))

        # Calculates acceptance rate for that temperature value
        acceptRate = nAccepted / AnnealingSchedule['nPerturb']
        tmp2 = pd.DataFrame({'T': [T], 'acceptRate': [acceptRate]})
        acceptRateChanges = acceptRateChanges.append(tmp2, ignore_index=True)
        print('Temperature: %.3f - Acceptance rate: %.2f - Cost: %.2f' % (T, acceptRate, E_m))

        # Stopping conditions
        if (acceptRate < AnnealingSchedule['StopAcceptanceRate']):
            LoopStopFlag = 1
            print('Optimization complete (acceptance rate threshold reached)')
        if (E_m <= AnnealingSchedule['StopCost']):
            LoopStopFlag = 1
            print('Optimization complete (cost objective reached)')

        # Update temperature
        T = T * AnnealingSchedule['ReducFactor']  # decrease temperature
        Tidx += 1  # next temperature step

    end = time.time()
    elapsedTime = end - start
    return R, Rchanges, acceptRateChanges, Cost, elapsedTime

def plotOptimizationResults(outdir, nReceivers, Rchanges, Cost, acceptRateChanges, R, iteridx=1):

    # plot Parameters evolution with Temperature
    f1 = plt.figure(1)
    for Ridx in range(nReceivers):
        plt.subplot(nReceivers,1, Ridx+1)
        plt.plot(Rchanges[Ridx,0,:], label='X(m)', color='black')
        plt.plot(Rchanges[Ridx,1,:], label='Y(m)', color='red')
        plt.plot(Rchanges[Ridx,2,:], label='Z(m)', color='green')
        plt.grid(True)
        plt.ylabel('H' + str(Ridx+1) )
        if Ridx == nReceivers -1:
            plt.xlabel('Temperature step')
        #if Ridx == 0:
        #    plt.legend(loc="best", labels=['X(m)','Y(m)','Z(m)'], bbox_to_anchor=(0.5,-0.1))
    f1.savefig(os.path.join(outdir, 'ReceiversPositionVsTemperature' + '_iteration-' + str(iteridx+1) + '.png'), bbox_inches='tight')

    # plot cost evolution with Temperature
    f2 = plt.figure(2)
    plt.plot(Cost['cost'], color = 'black')
    plt.grid(True)
    plt.xlabel('Temperature step')
    plt.ylabel('Cost')
    f2.savefig(os.path.join(outdir, 'CostVsTemperature' + '_iteration-' + str(iteridx+1) + '.png'), bbox_inches='tight')


    # plot acceptance rate with Temperature
    f3 = plt.figure(3)
    plt.semilogx(acceptRateChanges['T'],acceptRateChanges['acceptRate'], color = 'black')
    plt.grid(True)
    plt.xlabel('Temperature')
    plt.ylabel('Acceptance rate')
    plt.semilogx
    f3.savefig(os.path.join(outdir, 'AcceptanceRateVsTemperature' + '_iteration-' + str(iteridx+1) + '.png'), bbox_inches='tight')


    # plot Final receivers positions
    f4 = plt.figure(4)
    ax1 = f4.add_subplot(111, projection='3d')
    # Receivers
    ax1.scatter(R['x'], R['y'], R['z'], s=30, c='black')
    # Axes labels
    ax1.set_xlabel('X (m)', labelpad=10)
    ax1.set_ylabel('Y (m)', labelpad=10)
    ax1.set_zlabel('Z (m)', labelpad=10)
    plt.show()
    f4.savefig(os.path.join(outdir, 'FinalReceiversPosition' + '_iteration-' + str(iteridx+1) + '.png'), bbox_inches='tight')


def euclidean_dist(df1, df2, cols=['x', 'y', 'z']):
    """
    Calculate euclidean distance between two Pandas dataframes.

    Parameters
    ----------
    df1 : TYPE
        DESCRIPTION.
    df2 : TYPE
        DESCRIPTION.
    cols : TYPE, optional
        DESCRIPTION. The default is ['x','y','z'].

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    return np.linalg.norm(df1[cols].values - df2[cols].values, axis=0)


def calc_hydrophones_distances(hydrophones_coords):
    """
    Calculate Euclidiean distance between each hydrophone of an array.

    Parameters
    ----------
    hydrophones_coords : TYPE
        DESCRIPTION.

    Returns
    -------
    hydrophones_dist_matrix : TYPE
        DESCRIPTION.

    """
    hydrophones_dist_matrix = np.empty((len(hydrophones_coords),len(hydrophones_coords)))
    for index1, row1 in hydrophones_coords.iterrows():
        for index2, row2 in hydrophones_coords.iterrows():
            dist = euclidean_dist(row1, row2)
            hydrophones_dist_matrix[index1, index2] = dist
    return hydrophones_dist_matrix


def calc_tdoa(waveform_stack, hydrophone_pairs, sampling_frequency, TDOA_max_sec=None, upsample_res_sec=None, normalize=False, hydrophones_name=None, doplot=False):
    """
    TDOA measurements

    Calculates the time-difference of orrival (TDOA) between signals from
    different hydrophones by cross-correlation.

    Parameters
    ----------
    waveform_stack : list of numpy arrays
        Wavforms with amplitude values of the signal for each hydrophone.
        Each wavform is a numpy array which are stored in a list
        e.g. waveform_stack[0] contains a numpy array with the wavform from
         the first hydrophone.
    hydrophone_pairs : list
        Defines the pair of hydrophones for the TDOA measurements. Each element
        of hydrophones_pairs is a list with index values of the hydrophone in
        waveform_stack.
        e.g. hydrophones_pairs = [[3, 0], [3, 1], [3, 2], [3, 4], [3, 5]].
    sampling_frequency : float
        Sampling frequency of the waveform signals in  waveform_stack in Hz.
    TDOA_max_sec : float, optional
        Restricts the TDOA search to TDOA_max_sec seconds. The default is None.
    upsample_res_sec : float, optional
        If set, upsamples the wavforms in waveform_stack before the cross-
        correlation to have a time resolution of upsample_res_sec seconds.
        The default is None.
    normalize : bool, optional
        If set to True, normalizes the wavforms in waveform_stack to have a
        maximum amplitude of 1. The default is False.
    hydrophones_name : list, optional
        list of string with the name of each hydrophone. Only used for plots.
        The default is None.
    doplot : bool, optional
        If set to True, displays cross correlation plots for each hydrophone
        pair. The default is False.

    Returns
    -------
    tdoa_sec : 1-D numpy array
        Time-difference of arrival in seconds, for each hydrophone pair.
    tdoa_corr : 1-D numpy array
        Maximum cross-correlation value for each hydrophone pair (between 0
        and 1).

    """
    tdoa_sec = []
    tdoa_corr = []
    # Upsampling
    if upsample_res_sec:
        if upsample_res_sec < (1/sampling_frequency):
            for chan_id, waveform in enumerate(waveform_stack):
                waveform_stack[chan_id], new_sampling_frequency = upsample(
                    waveform, 1/sampling_frequency, upsample_res_sec)
            sampling_frequency = new_sampling_frequency
        else:
            print('Warning: upsampling not applied because the requested time'
                  ' resolution (upsample_res_sec) is larger than the current'
                  ' time resolution of the signal.')
    # Normalize max amplitude to 1
    if normalize:
        for chan_id, waveform in enumerate(waveform_stack):
                    waveform_stack[chan_id] = waveform / np.max(waveform)
    # Constrains to a max TDOA (based on array geometry)
    if TDOA_max_sec:
        TDOA_max_samp = int(np.round(TDOA_max_sec*sampling_frequency))

    # cross correlation
    for hydrophone_pair in hydrophone_pairs:
        # signal from each hydrophone
        s1 = waveform_stack[hydrophone_pair[0]]
        s2 = waveform_stack[hydrophone_pair[1]]
        # cross correlation
        corr = scipy.signal.correlate(s2,s1, mode='full', method='auto')
        corr = corr/(np.linalg.norm(s1)*np.linalg.norm(s2))
        lag_array = scipy.signal.correlation_lags(s2.size,s1.size, mode="full")
        # Identify correlation peak within the TDOA search window (SW)
        if TDOA_max_sec:
            SW_start_idx = np.where(lag_array == -TDOA_max_samp)[0][0] # search window start idx
            SW_stop_idx = np.where(lag_array == TDOA_max_samp)[0][0] # search window stop idx
        else:
            SW_start_idx=0
            SW_stop_idx=len(corr)-1
        corr_max_idx = np.argmax(corr[SW_start_idx:SW_stop_idx]) + SW_start_idx # idx of max corr value
        delay = lag_array[corr_max_idx]
        corr_value = corr[corr_max_idx]
        tdoa_sec.append(delay/sampling_frequency)
        tdoa_corr.append(corr_value)

        if doplot:
            fig, ax = plt.subplots(nrows=2, sharex=False)
            if hydrophones_name:
                label1= hydrophones_name[hydrophone_pair[0]] + ' (ref)'
                label2= hydrophones_name[hydrophone_pair[1]]
            else:
                 label1 = 'Hydrophone ' + str(hydrophone_pair[0]) + ' (ref)'
                 label2 = 'Hydrophone ' + str(hydrophone_pair[1])
            ax[0].plot(s1, color='red', label=label1)
            ax[0].plot(s2, color='black',label=label2)
            ax[0].set_xlabel('Time (sample)')
            ax[0].set_ylabel('Amplitude')
            ax[0].legend()
            ax[0].grid()
            ax[0].set_title('TDOA: ' + str(delay) + ' samples')
            ax[1].plot(lag_array, corr)
            ax[1].plot(delay, corr_value ,marker = '.', color='r', label='TDOA')
            if TDOA_max_sec:
                width = 2*TDOA_max_samp
                height = 2
                rect = plt.Rectangle((-TDOA_max_samp, -1), width, height,
                                             linewidth=1,
                                             edgecolor='green',
                                             facecolor='green',
                                             alpha=0.3,
                                             label='Search window')
                ax[1].add_patch(rect)
            ax[1].set_xlabel('Lag (sample)')
            ax[1].set_ylabel('Correlation')
            ax[1].set_title('Correlation: ' + str(corr_value))
            ax[1].set_ylim(-1,1)
            ax[1].grid()
            ax[1].legend()
            plt.tight_layout()
    return np.array([tdoa_sec]).transpose(), np.array([tdoa_corr]).transpose()


def solve_iterative_ML(d, hydrophones_coords, hydrophone_pairs, m, V, damping_factor):
    # Define the Jacobian matrix: Eq. (5) in Mouy et al. (2018)
    A = defineJacobian(hydrophones_coords, m, V, hydrophone_pairs)
    # Predicted TDOA at m (forward problem): Eq. (1) in Mouy et al. (2018)
    d0 = predict_tdoa(m, V, hydrophones_coords, hydrophone_pairs)
    # Reformulation of the problem
    delta_d = d-d0 # Delta d: measured data - predicted data
    try:
        # Resolving by creeping approach(ML inverse for delta m): Eq. (6) in Mouy et al. (2018)
        delta_m = np.dot(np.dot(np.linalg.inv(np.dot(A.transpose(),A)),A.transpose()),delta_d) #general inverse
        # New model: Eq. (7) in Mouy et al. (2018)
        m_new = m.iloc[0].values + (damping_factor*delta_m.transpose())
        m['x'] = m_new[0,0]
        m['y'] = m_new[0,1]
        m['z'] = m_new[0,2]
        # ## jumping approach
        # #m=inv(A'*CdInv*A)*A'*CdInv*dprime; % retrieved model using ML
        # Data misfit
        part1 = np.dot(A,delta_m)-delta_d
        data_misfit = np.dot(part1.transpose(), part1)
    except np.linalg.LinAlgError as err:
        if 'Singular matrix' in str(err):
            print('Error when inverting for delta_m: Singular Matrix')
            m['x'] = np.nan
            m['y'] = np.nan
            m['z'] = np.nan
            data_misfit = np.array([[np.nan]])
        else:
            raise err

    return m, data_misfit[0, 0]


def linearized_inversion(d, hydrophones_coords,hydrophone_pairs,inversion_params, sound_speed_mps, doplot=False):
    # convert parameters to numpy arrays
    damping_factor = np.array(inversion_params['damping_factor'])
    Tdelta_m = np.array(inversion_params['stop_delta_m'])
    V = np.array(sound_speed_mps)
    max_iteration = np.array(inversion_params['stop_max_iteration'])
    # define starting model(s)
    start_models = [inversion_params['start_model']]
    if inversion_params['start_model_repeats'] > 1:
        x_bounds = [hydrophones_coords['x'].min(), hydrophones_coords['x'].max()]
        y_bounds = [hydrophones_coords['y'].min(), hydrophones_coords['y'].max()]
        z_bounds = [hydrophones_coords['z'].min(), hydrophones_coords['z'].max()]
        for m_nb in range(1, inversion_params['start_model_repeats']):
            x_tmp = np.random.uniform(low=x_bounds[0], high=x_bounds[1])
            y_tmp = np.random.uniform(low=y_bounds[0], high=y_bounds[1])
            z_tmp = np.random.uniform(low=z_bounds[0], high=z_bounds[1])
            start_models.append([x_tmp, y_tmp, z_tmp])
    m_stack = []
    iterations_logs_stack = []
    data_misfit_stack = []
    for start_model in start_models:
        # current starting model
        m = pd.DataFrame({'x': [start_model[0]], 'y': [start_model[1]], 'z': [start_model[2]]})
        # Keeps track of values for each iteration
        iterations_logs = pd.DataFrame({
            'x': [start_model[0]],
            'y': [start_model[1]],
            'z': [start_model[2]],
            'norm': np.nan,
            'data_misfit': np.nan,
            })
        # Start iterations
        stop = False
        idx=0
        while stop == False:
            idx=idx+1
            # Linear inversion
            m_it, data_misfit_it = solve_iterative_ML(d, hydrophones_coords, hydrophone_pairs, m, V, damping_factor)
            # Save model and data misfit for each iteration
            iterations_logs = iterations_logs.append({
                'x': m_it['x'].values[0],
                'y': m_it['y'].values[0],
                'z': m_it['z'].values[0],
                #'norm': np.sqrt(np.square(m_it).sum(axis=1)).values[0],
                'norm': np.linalg.norm(m_it.iloc[0].to_numpy() - iterations_logs.iloc[-1][['x','y','z']].to_numpy(),np.inf),
                'data_misfit': data_misfit_it,
                }, ignore_index=True)
            # Update m
            m = m_it
            # stopping criteria
            if idx>1:
                 # norm of model diference
                #if (iterations_logs['norm'][idx] - iterations_logs['norm'][idx-1] <= Tdelta_m):
                if iterations_logs['norm'][idx] <= Tdelta_m:
                    stop = True
                    if np.isnan(m['x'].values[0]):
                        converged=False # Due to singular matrix
                        print('Singular matrix - inversion hasn''t converged.')
                    else:
                        converged=True
                elif idx > max_iteration: # max iterations exceeded
                    stop = True
                    converged=False
                    print('Max iterations reached - inversion hasn''t converged.')
                # elif iterations_logs['norm'][idx] > iterations_logs['norm'][idx-1]: # if norm starts increasing -> then stop there
                #     stop = True
                #     converged=False
                #     print('Norm increasing - inversion hasn''t converged.')
                elif np.isnan(m['x'].values[0]):
                    stop = True
                    converged=False
                    print('Singular matrix - inversion hasn''t converged.')

        if not converged:
            m['x']=np.nan
            m['y']=np.nan
            m['z']=np.nan
        # Save results for each starting model
        m_stack.append(m)
        iterations_logs_stack.append(iterations_logs)
        data_misfit_stack.append(iterations_logs['data_misfit'].iloc[-1])
    # Select results for starting model with best ending data misfit
    min_data_misfit = np.nanmin(data_misfit_stack)
    if np.isnan(min_data_misfit):
        min_idx=0
    else:
        min_idx = data_misfit_stack.index(min_data_misfit)
    return m_stack[min_idx], iterations_logs_stack[min_idx]


# -*- coding: utf-8 -*-
"""
Created on Wed Apr  7 10:07:53 2021

@author: xavier.mouy
"""
import os
import sys
sys.path.append(r'C:\Users\xavier.mouy\Documents\GitHub\ecosound') # Adds higher directory to python modules path.

import pandas as pd
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d
import matplotlib.cm
#import scipy.spatial
import numpy as np
import datetime
import math

from ecosound.core.audiotools import Sound, upsample
from ecosound.core.spectrogram import Spectrogram
from ecosound.core.measurement import Measurement
from ecosound.detection.detector_builder import DetectorFactory
from ecosound.visualization.grapher_builder import GrapherFactory
import ecosound.core.tools
from ecosound.core.tools import derivative_1d, envelope, read_yaml
from localizationlib import euclidean_dist, calc_hydrophones_distances, calc_tdoa, defineReceiverPairs, defineJacobian, predict_tdoa, linearized_inversion, solve_iterative_ML, defineCubeVolumeGrid, defineSphereVolumeGrid
import platform

def find_audio_files(filename, hydrophones_config):
    """ Find corresponding files and channels for all the hydrophones of the array """
    filename = os.path.basename(infile)
    # Define file tail
    for file_root in hydrophones_config['file_name_root']:
        idx = filename.find(file_root)
        if idx >= 0:
            file_tail = filename[len(file_root):]
            break
    # Loop through channels and define all files paths and audio channels
    audio_files = {'path':[], 'channel':[]}
    for row_idx, row_data in hydrophones_config.iterrows():
        file_path = os.path.join(row_data.data_path, row_data.file_name_root + file_tail)
        chan = row_data.audio_file_channel
        audio_files['path'].append(file_path)
        audio_files['channel'].append(chan)
    return audio_files

def run_detector(infile, channel, config, chunk=None, deployment_file=None):

    sound = Sound(infile)
    # load audio data
    if chunk:
        sound.read(channel=channel, chunk=chunk, unit='sec', detrend=True)
        time_offset_sec = chunk[0]
    else:
        sound.read(channel=channel, detrend=True)
        time_offset_sec = 0

    # Calculates  spectrogram
    spectro = Spectrogram(config['SPECTROGRAM']['frame_sec'],
                                  config['SPECTROGRAM']['window_type'],
                                  config['SPECTROGRAM']['nfft_sec'],
                                  config['SPECTROGRAM']['step_sec'],
                                  sound.waveform_sampling_frequency,
                                  unit='sec',)
    spectro.compute(sound,
                    config['SPECTROGRAM']['dB'],
                    config['SPECTROGRAM']['use_dask'],
                    config['SPECTROGRAM']['dask_chunks'],)

    spectro.crop(frequency_min=config['SPECTROGRAM']['fmin_hz'],
                         frequency_max=config['SPECTROGRAM']['fmax_hz'],
                         inplace=True,
                         )
    # Denoise
    spectro.denoise(config['DENOISER']['denoiser_name'],
                    window_duration=config['DENOISER']['window_duration_sec'],
                    use_dask=config['DENOISER']['use_dask'],
                    dask_chunks=tuple(config['DENOISER']['dask_chunks']),
                    inplace=True)
    # Detector
    file_timestamp = ecosound.core.tools.filename_to_datetime(sound.file_full_path)[0]
    detector = DetectorFactory(config['DETECTOR']['detector_name'],
                               kernel_duration=config['DETECTOR']['kernel_duration_sec'],
                               kernel_bandwidth=config['DETECTOR']['kernel_bandwidth_hz'],
                               threshold=config['DETECTOR']['threshold'],
                               duration_min=config['DETECTOR']['duration_min_sec'],
                               bandwidth_min=config['DETECTOR']['bandwidth_min_hz']
                               )
    start_time = file_timestamp + datetime.timedelta(seconds=time_offset_sec)
    detections = detector.run(spectro,
                              start_time=start_time,
                              use_dask=config['DETECTOR']['use_dask'],
                              dask_chunks=tuple(config['DETECTOR']['dask_chunks']),
                              debug=False,
                              )
    # add time offset in only a section of recording was analysed.
    detections.data['time_min_offset'] = detections.data['time_min_offset'] + time_offset_sec
    detections.data['time_max_offset'] = detections.data['time_max_offset'] + time_offset_sec

    # add deployment metadata
    detections.insert_metadata(deployment_file, channel=channel)

    # Add file informations
    file_name = os.path.splitext(os.path.basename(sound.file_full_path))[0]
    file_dir = os.path.dirname(sound.file_full_path)
    file_ext = os.path.splitext(sound.file_full_path)[1]
    detections.insert_values(operator_name=platform.uname().node,
                               audio_file_name=file_name,
                               audio_file_dir=file_dir,
                               audio_file_extension=file_ext,
                               audio_file_start_date=ecosound.core.tools.filename_to_datetime(sound.file_full_path)[0]
                               )

    return detections


def plot_data(audio_files,frame, window_type, nfft, step, fmin, fmax, chunk = None, detections=None, detections_channel=0):
    graph_spectros = GrapherFactory('SoundPlotter', title='Spectrograms', frequency_max=fmax)
    graph_waveforms = GrapherFactory('SoundPlotter', title='Waveforms')
    for audio_file, channel in zip(audio_files['path'], audio_files['channel'] ): # for each channel
        # load waveform
        sound = Sound(audio_file)
        sound.read(channel=channel, chunk=chunk, unit='sec', detrend=True)
        # Calculates  spectrogram
        spectro = Spectrogram(frame, window_type, nfft, step, sound.waveform_sampling_frequency, unit='sec')
        spectro.compute(sound, dB=True, use_dask=False)
        # Crop unused frequencies
        spectro.crop(frequency_min=fmin, frequency_max=fmax, inplace=True)
        # Plot
        graph_spectros.add_data(spectro, time_offset_sec=chunk[0])
        graph_waveforms.add_data(sound, time_offset_sec=chunk[0])

    graph_spectros.colormap = 'binary'
    if detections:
        graph_spectros.add_annotation(detections, panel=detections_channel, color='green',label='Detections')
        graph_waveforms.add_annotation(detections, panel=detections_channel, color='green',label='Detections')

    if chunk:
        graph_spectros.time_min = chunk[0]
        graph_spectros.time_max = chunk[1]
        graph_waveforms.time_min = chunk[0]
        graph_waveforms.time_max = chunk[1]

    graph_spectros.show()
    graph_waveforms.show()


def stack_waveforms(audio_files, detec, TDOA_max_sec):
    waveform_stack = []
    for audio_file, channel in zip(audio_files['path'], audio_files['channel'] ): # for each channel
        # load waveform
        chan_wav = Sound(audio_file)
        chan_wav.read(channel=channel,
                      chunk=[detec['time_min_offset']-TDOA_max_sec, detec['time_max_offset']+TDOA_max_sec],
                      unit='sec',
                      detrend=True)
        # bandpass filter
        chan_wav.filter('bandpass', [detec['frequency_min'], detec['frequency_max']])
        # stack
        waveform_stack.append(chan_wav.waveform)
    return waveform_stack


def calc_data_error(tdoa_sec, m, sound_speed_mps,hydrophones_config, hydrophone_pairs):
    """ Calculates tdoa measurement errors. Eq. (9) in Mouy et al. 2018"""
    tdoa_m = predict_tdoa(m, sound_speed_mps, hydrophones_config, hydrophone_pairs)
    Q = len(m)
    M = m.size
    N = len(tdoa_sec)
    error_std = np.sqrt((1/(Q*(N-M))) * (sum((tdoa_sec-tdoa_m)**2)))
    return error_std


def calc_loc_errors(tdoa_errors_std, m, sound_speed_mps, hydrophones_config, hydrophone_pairs):
    """ Calculates localization errors. Eq. (8) in Mouy et al. 2018."""
    A = defineJacobian(hydrophones_config, m, sound_speed_mps, hydrophone_pairs)
    Cm = (tdoa_errors_std**2) * np.linalg.inv(np.dot(A.transpose(),A)) # Model covariance matrix for IID
    err_std = np.sqrt(np.diag(Cm))
    return pd.DataFrame({'x_std': [err_std[0]], 'y_std': [err_std[1]], 'z_std': [err_std[2]]})

def cartesian2spherical (x,y,z):
    # Converting cartesian to polar coordinate
    const = 180/np.pi
    XsqPlusYsq = x**2 + y**2
    r = math.sqrt(XsqPlusYsq + z**2)               # r
    elev = math.atan2(math.sqrt(XsqPlusYsq),z)     # theta
    az = math.atan2(y,x)*180/np.pi
    return r, az, elev

# # Calculating radius
# radius = math.sqrt( x * x + y * y )
# # Calculating angle (theta) in radian
# theta = math.atan(y/x)
# # Converting theta from radian to degree
# theta = 180 * theta/math.pi

def run_localization(infile, deployment_info_file, detection_config, hydrophones_config,localization_config):
    t1 = 0
    t2 = 70   
    # Look up data files for all channels
    audio_files = find_audio_files(infile, hydrophones_config)

    # run detector on selected channel
    print('DETECTION')
    detections = run_detector(audio_files['path'][detection_config['AUDIO']['channel']],
                              audio_files['channel'][detection_config['AUDIO']['channel']],
                              detection_config,
                              chunk = [t1, t2],
                              deployment_file=deployment_info_file)
    #detections.insert_values(frequency_min=20)
    
    print(str(len(detections)) + ' detections')
    
    # # plot spectrogram/waveforms of all channels and detections
    # plot_data(audio_files,
    #           detection_config['SPECTROGRAM']['frame_sec'],
    #           detection_config['SPECTROGRAM']['window_type'],
    #           detection_config['SPECTROGRAM']['nfft_sec'],
    #           detection_config['SPECTROGRAM']['step_sec'],
    #           detection_config['SPECTROGRAM']['fmin_hz'],
    #           detection_config['SPECTROGRAM']['fmax_hz'],
    #           chunk = [t1, t2],
    #           detections=detections,
    #           detections_channel=detection_config['AUDIO']['channel'])
    
    # localization
    sound_speed_mps = localization_config['ENVIRONMENT']['sound_speed_mps']
    ref_channel = localization_config['TDOA']['ref_channel']
    
    # define search window based on hydrophone separation and sound speed
    hydrophones_dist_matrix = calc_hydrophones_distances(hydrophones_config)
    TDOA_max_sec = np.max(hydrophones_dist_matrix)/sound_speed_mps
    
    # define hydrophone pairs
    hydrophone_pairs = defineReceiverPairs(len(hydrophones_config), ref_receiver=ref_channel)
    
    # pre-compute grid search if needed
    if localization_config['METHOD']['grid_search']:
        sources = defineSphereVolumeGrid(
            localization_config['GRIDSEARCH']['spacing_m'],
            localization_config['GRIDSEARCH']['radius_m'],
            origin=localization_config['GRIDSEARCH']['origin'])
        #sources = defineCubeVolumeGrid(0.2, 2, origin=[0, 0, 0])
        sources_tdoa = np.zeros(shape=(len(hydrophone_pairs),len(sources)))
        for source_idx, source in sources.iterrows():
            sources_tdoa[:,source_idx] = predict_tdoa(source, sound_speed_mps, hydrophones_config, hydrophone_pairs).T
        theta = np.arctan2(sources['y'].to_numpy(),sources['x'].to_numpy())*(180/np.pi) # azimuth
        phi = np.arctan2(sources['y'].to_numpy()**2+sources['x'].to_numpy()**2,sources['z'].to_numpy())*(180/np.pi)
        sources['theta'] = theta
        sources['phi'] = phi
    
    # Define Measurement object for the localization results
    if localization_config['METHOD']['linearized_inversion']:
        localizations = Measurement()
        localizations.metadata['measurer_name'] = localization_method_name
        localizations.metadata['measurer_version'] = '0.1'
        localizations.metadata['measurements_name'] = [['x', 'y', 'z', 'x_std', 'y_std', 'z_std', 'tdoa_errors_std']]
    # need to define what output is for grid search
    
    # pick single detection (will use loop after)
    print('LOCALIZATION')
    for detec_idx, detec in detections.data.iterrows():
    
        if 'detec_idx_forced' in locals():
            print('Warning: forced to only process detection #', str(detec_idx_forced))
            detec = detections.data.iloc[detec_idx_forced]
        
        print( str(detec_idx+1) + '/' + str(len(detections)))
    
        # load data from all channels for that detection
        waveform_stack = stack_waveforms(audio_files, detec, TDOA_max_sec)
    
        # readjust signal boundaries to only focus on section with most energy 
        percentage_max_energy = 90
        chunk = ecosound.core.tools.tighten_signal_limits_peak(waveform_stack[detection_config['AUDIO']['channel']], percentage_max_energy)
        waveform_stack = [x[chunk[0]:chunk[1]] for x in waveform_stack]
    
        # calculate TDOAs
        tdoa_sec, corr_val = calc_tdoa(waveform_stack,
                                       hydrophone_pairs,
                                       detec['audio_sampling_frequency'],
                                       TDOA_max_sec=TDOA_max_sec,
                                       upsample_res_sec=localization_config['TDOA']['upsample_res_sec'],
                                       normalize=localization_config['TDOA']['normalize'],
                                       doplot=False,
                                       )
    
        if localization_config['METHOD']['grid_search']:    
            delta_tdoa = sources_tdoa - tdoa_sec
            delta_tdoa_norm = np.linalg.norm(delta_tdoa, axis=0)
            sources['delta_tdoa'] = delta_tdoa_norm
        
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            colors = matplotlib.cm.tab10(hydrophones_config.index.values)
            #alphas = delta_tdoa_norm - min(delta_tdoa_norm)
            #alphas = alphas/max(alphas)
            #alphas = alphas - 1
            #alphas = abs(alphas)
            #alphas = np.array(alphas)
            alphas = 0.5
            for index, hp in hydrophones_config.iterrows():
                point = ax.scatter(hp['x'],hp['y'],hp['z'],
                                s=40,
                                color=colors[index],
                                label=hp['name'],
                                )
            ax.scatter(sources['x'],
                        sources['y'],
                        sources['z'],
                        c=sources['delta_tdoa'],
                        s=2,
                        alpha=alphas,)
            # Axes labels
            ax.set_xlabel('X (m)', labelpad=10)
            ax.set_ylabel('Y (m)', labelpad=10)
            ax.set_zlabel('Z (m)', labelpad=10)
            # legend
            ax.legend(bbox_to_anchor=(1.07, 0.7, 0.3, 0.2), loc='upper left')
            plt.tight_layout()
            plt.show()
        
            plt.figure()
            sources.plot.hexbin(x="theta",
                                y="phi",
                                C="delta_tdoa",
                                reduce_C_function=np.mean,
                                gridsize=40,
                                cmap="viridis")
        
        # Lineralized inversion
        if localization_config['METHOD']['linearized_inversion']:
            [m, iterations_logs] = linearized_inversion(tdoa_sec,
                                                        hydrophones_config,
                                                        hydrophone_pairs,
                                                        localization_config['INVERSION'],
                                                        sound_speed_mps,
                                                        doplot=False)
        
            # Estimate uncertainty
            tdoa_errors_std = calc_data_error(tdoa_sec, m, sound_speed_mps,hydrophones_config, hydrophone_pairs)
            loc_errors_std = calc_loc_errors(tdoa_errors_std, m, sound_speed_mps, hydrophones_config, hydrophone_pairs)
    
        # Bring all detection and localization informations together
        detec.loc['x'] = m['x'].values[0]
        detec.loc['y'] = m['y'].values[0]
        detec.loc['z'] = m['z'].values[0]
        detec.loc['x_std'] = loc_errors_std['x_std'].values[0]
        detec.loc['y_std'] = loc_errors_std['y_std'].values[0]
        detec.loc['z_std'] = loc_errors_std['z_std'].values[0]
        detec.loc['tdoa_errors_std'] = tdoa_errors_std[0]
    
        # stack to results into localization object
        localizations.data = localizations.data.append(detec, ignore_index=True)
        
    return localizations



## ############################################################################
indir = r'C:\Users\xavier.mouy\Documents\Reports_&_Papers\Papers\10-XAVarray_2020\data\large_array\2019-09-15_HornbyIsland_AMAR_07-HI\test'
outdir=r'C:\Users\xavier.mouy\Documents\GitHub\ecosound\ecosound\localization'

deployment_info_file = r'C:\Users\xavier.mouy\Documents\Reports_&_Papers\Papers\10-XAVarray_2020\data\large_array\2019-09-15_HornbyIsland_AMAR_07-HI\deployment_info.csv'
hydrophones_config_file = r'C:\Users\xavier.mouy\Documents\Reports_&_Papers\Papers\10-XAVarray_2020\data\large_array\2019-09-15_HornbyIsland_AMAR_07-HI\hydrophones_config_07-HI.csv'
detection_config_file = r'C:\Users\xavier.mouy\Documents\Reports_&_Papers\Papers\10-XAVarray_2020\config_files\detection_config_large_array_HF2.yaml'
localization_config_file = r'C:\Users\xavier.mouy\Documents\Reports_&_Papers\Papers\10-XAVarray_2020\config_files\localization_config_large_array.yaml'



# load configuration parameters
hydrophones_config= pd.read_csv(hydrophones_config_file, skipinitialspace=True, dtype={'name': str, 'file_name_root': str}) # load hydrophone coordinates (meters)
detection_config = read_yaml(detection_config_file)
localization_config = read_yaml(localization_config_file)
if localization_config['METHOD']['linearized_inversion']:
    localization_method_name = 'Linearized inversion'
    if localization_config['METHOD']['grid_search']:
        raise ValueError('Only 1 localization method allowed.')
else:
    localization_method_name = 'Grid search'
    if localization_config['METHOD']['grid_search'] == False:
        raise ValueError('At least 1 localization method needs to be defined.')


files = ecosound.core.tools.list_files(indir, '.wav',
                                       recursive=False,
                                       case_sensitive=True,
                                       )

nfiles = len(files) 
for idx, infile in enumerate(files):    
    print(idx+1, '/', nfiles, os.path.split(infile)[1])      
    outfile = os.path.join(outdir, os.path.split(infile)[1] + '.nc')   
    if os.path.exists(outfile) is False: 
        loc = run_localization(infile, deployment_info_file, detection_config, hydrophones_config,localization_config)
        loc.to_netcdf(outfile)
    else:
        print('File already processed')
    print('s')






# -*- coding: utf-8 -*-
"""
Created on Thu Sep  7 15:26:32 2017

@author: xavier
"""

import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d
import plotly
import plotly.graph_objs as go
import pickle

  

picklefile=r'C:\Users\xavier.mouy\Documents\Workspace\GitHub\Fish-localization\results\ArrayOptimization\20180711093903_Receivers5_Bounds1m_Sources500_Radius2m\ArrayOptimizationResults_iteration-2.pickle'

data = pickle.load(open(picklefile, 'rb'))
outroot = data["outroot"]
outdir = data["outdir"]
nIter = data["nIter"]
nsources = data["nsources"]
radius = data["radius"]
origin = data["origin"]
spacing = data["spacing"]
NoiseSTD = data["NoiseSTD"]
V = data["V"]
nReceivers = data["nReceivers"]
ReceiverBoundValue = data["ReceiverBoundValue"]
ReceiverBounds = data["ReceiverBounds"]
AnnealingSchedule = data["AnnealingSchedule"]
R = data["R"]
Rchanges = data["Rchanges"]
acceptRateChanges = data["acceptRateChanges"]
Cost = data["Cost"]
S = data["S"]
Uncertainties2 = data["Uncertainties2"]
                  

nReceivers = R.shape[0]

# plot Parameters evolution with Temperature
f1 = plt.figure(1)
for Ridx in range(nReceivers):
    plt.subplot(nReceivers,1, Ridx+1)
    plt.plot(Rchanges[Ridx,0,:], label='X(m)', color='black')
    plt.plot(Rchanges[Ridx,1,:], label='Y(m)', color='red')
    plt.plot(Rchanges[Ridx,2,:], label='Z(m)', color='green')
    plt.grid('on')
    plt.ylabel('H' + str(Ridx+1) )
    if Ridx == nReceivers -1:
        plt.xlabel('Temperature step')
    #if Ridx == 0:
    #    plt.legend(loc="best", labels=['X(m)','Y(m)','Z(m)'], bbox_to_anchor=(0.5,-0.1))

# plot cost evolution with Temperature
f2 = plt.figure(2)
plt.plot(Cost['cost'], color = 'black')
plt.grid('on')
plt.xlabel('Temperature step')
plt.ylabel('Cost')

# plot acceptance rate with Temperature
f3 = plt.figure(3)
plt.semilogx(acceptRateChanges['T'],acceptRateChanges['acceptRate'], color = 'black')
plt.grid('on')
plt.xlabel('Temperature')
plt.ylabel('Acceptance rate')
plt.semilogx

# plot Parameters evolution with Temperature
f4 = plt.figure(4)
ax1 = f4.add_subplot(111, projection='3d')
# Receivers
ax1.scatter(R['x'], R['y'], R['z'], s=30, c='black')
# Axes labels
ax1.set_xlabel('X (m)', labelpad=10)
ax1.set_ylabel('Y (m)', labelpad=10)
ax1.set_zlabel('Z (m)', labelpad=10)
plt.show()




#f1.legend(loc="upper right", labels=['X(m)','Y(m)','Z(m)'], bbox_to_anchor=(0.5,-0.1))
trace = go.Scatter3d(
        x=R['x'],
        y=R['y'],
        z=R['z'],
        mode='markers',
        marker=dict(size=10,color='red',colorscale='Viridis'),
        text= ['H1','H2','H3','H4','H5','H6']
    )


data = [trace]
plotly.offline.plot(data,show_link=False,config=dict(displayModeBar=False), filename='OptimalReceiversLocation.html')




trace1 = go.Scatter3d(
        x=S['x'],
        y=S['y'],
        z=S['z'],
        mode='markers',
        marker=dict(size=8,color=Uncertainties2['rms'],colorscale='Viridis'),    
        text = Uncertainties2['rms']
    )


data = [trace1, trace]
plotly.offline.plot(data, show_link=False,config=dict(displayModeBar=False), filename='OptimalReceiversUncertainties.html' )




# -*- coding: utf-8 -*-
"""
Created on Mon Sep 20 14:55:40 2021

@author: xavier.mouy
"""
import os
import sys
sys.path.append(r'C:\Users\xavier.mouy\Documents\GitHub\ecosound') # Adds higher directory to python modules path.

import pandas as pd
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d
import matplotlib.cm
import matplotlib as mpl
from mpl_toolkits.axes_grid1 import make_axes_locatable
#import scipy.spatial
import numpy as np
import datetime
import math
import cv2

from ecosound.core.audiotools import Sound, upsample
from ecosound.core.spectrogram import Spectrogram
from ecosound.core.measurement import Measurement
from ecosound.detection.detector_builder import DetectorFactory
from ecosound.visualization.grapher_builder import GrapherFactory
import ecosound.core.tools
from ecosound.core.tools import derivative_1d, envelope, read_yaml
from localizationlib import euclidean_dist, calc_hydrophones_distances, calc_tdoa, defineReceiverPairs, defineJacobian, predict_tdoa, linearized_inversion, solve_iterative_ML, defineCubeVolumeGrid, defineSphereVolumeGrid
import platform
import cv2

def plot_spectrogram(audio_file,loc,t1_sec, t2_sec, geometry=(1,1,1)):
    
    fmin=0
    fmax=1000
    frame= 0.0625
    window_type= 'hann'
    nfft=0.0853
    step=0.01
    channel=0
    chunk=[t1_sec,t2_sec]
    
    graph_spectros = GrapherFactory('SoundPlotter', title='Spectrograms', frequency_max=fmax)
    sound = Sound(audio_file)
    sound.read(channel=channel, chunk=chunk, unit='sec', detrend=True)
    # Calculates  spectrogram
    spectro = Spectrogram(frame, window_type, nfft, step, sound.waveform_sampling_frequency, unit='sec')
    spectro.compute(sound, dB=True, use_dask=False)
    # Crop unused frequencies
    spectro.crop(frequency_min=fmin, frequency_max=fmax, inplace=True)
    # Plot
    graph_spectros.add_data(spectro)    
    
    #graph_spectros.add_annotation(loc, panel=0, color='burlywood',label='Detections')
    graph_spectros.add_annotation(loc, panel=0, color='peachpuff')
    
    graph_spectros.colormap = 'binary' #'jet'
    fig, ax = graph_spectros.show()

    if ax.get_geometry() != geometry :
        ax.change_geometry(*geometry)        
    return fig, ax

def plot_top_view(hydrophones_config,loc_data,params,cmap,norm, ax):
    
    #fig1 = plt.figure()
    #ax = fig1.add_subplot(111)
    colors = matplotlib.cm.tab10(hydrophones_config.index.values)
    
    # plot hydrophones
    for index, hp in hydrophones_config.iterrows():
        point = ax.scatter(hp['x'],hp['y'],
                           s=20,
                           #color=colors[index],
                           color='gainsboro',
                           edgecolors='dimgray',
                           #markeredgecolor='r',
                           label=hp['name'],
                           alpha=1,
                           zorder=3
                           )

        # plot frame
        # plot frame
        frame_color = 'whitesmoke'
        frame_alpha = 1
        frame_width = 3        
        # plot frame
        # plot fishcam 02

        ax.plot([-0.3,0.4],[-0.1,-0.1],
              linewidth=frame_width ,
              alpha=frame_alpha,
              linestyle= 'solid',
              color=frame_color,
              )

        rectangle3 = plt.Rectangle((-0.3, -0.7), 0.7, 0.9,linewidth=frame_width,ec=frame_color,alpha=frame_alpha, facecolor='white')        
        ax.add_patch(rectangle3)
        
        rectangle4 = plt.Rectangle((-0.04, -0.4), 0.22, 0.5,linewidth=frame_width,ec=frame_color,alpha=frame_alpha, facecolor=frame_color)
        ax.add_patch(rectangle4)
        
        ax.plot([-0.5,0.63],[0, 0],
              linewidth=1 ,
              alpha=frame_alpha,
              linestyle= 'solid',
              color='dimgray',
              )
        ax.plot([0.13, 0.13],[0,0.57],
              linewidth=1 ,
              alpha=frame_alpha,
              linestyle= 'solid',
              color='dimgray',
              )      


        
        
        # ax.plot([0.93, 1.5],[-0.76, -0.2],
        #      linewidth=frame_width ,
        #      alpha=frame_alpha,
        #      linestyle= 'solid',
        #      color=frame_color,
        #      )
        # ax.plot([1.5, 2],[-0.1, -1.06],
        #      linewidth=frame_width ,
        #      alpha=frame_alpha,
        #      linestyle= 'solid',
        #      color=frame_color,
        #      )
        

    # plot localizations
    ax.scatter(loc_data['x'], loc_data['y'],c=loc_data['time_min_offset'],
                        s=params['loc_size'].values[0],
                        marker=params['loc_marker'].values[0],
                        #color=params['loc_color'].values[0],
                        alpha=params['loc_alpha'].values[0],
                        cmap=cmap,
                        norm=norm,
                        zorder=5,
                        )
    # plot uncertainties
    for idx, loc_point in loc_data.iterrows():   
        ax.plot([loc_point['x_min_CI99'],loc_point['x_max_CI99']],
                [loc_point['y'],loc_point['y']],
                #c=loc_point['time_min_offset'],
                linewidth=params['uncertainty_width'].values[0],
                linestyle=params['uncertainty_style'].values[0],
                #color=params['uncertainty_color'].values[0],
                color=cmap(norm(loc_point['time_min_offset'])),
                alpha=params['uncertainty_alpha'].values[0],
                #cmap=cmap,
                #norm=norm,
                )
    
        ax.plot([loc_point['x'],loc_point['x']],
                [loc_point['y_min_CI99'],loc_point['y_max_CI99']],
                linewidth=params['uncertainty_width'].values[0],
                linestyle=params['uncertainty_style'].values[0],
                #color=params['uncertainty_color'].values[0],
                color=cmap(norm(loc_point['time_min_offset'])),
                alpha=params['uncertainty_alpha'].values[0],
                )
        
    # Axes labels
    ax.set_xlabel('X (m)', labelpad=4)
    ax.set_ylabel('Y (m)', labelpad=4)
    ax.set_xlim(params['x_min'].values[0], params['x_max'].values[0])
    ax.set_ylim(params['y_min'].values[0], params['y_max'].values[0])
    #ax.grid()
    ax.set_aspect('equal', adjustable='box')
    return ax


def plot_side_view(hydrophones_config,loc_data,params,cmap,norm, ax):
    
    #fig1 = plt.figure()
    #ax = fig1.add_subplot(111)
    colors = matplotlib.cm.tab10(hydrophones_config.index.values)
    
    # plot hydrophones
    for index, hp in hydrophones_config.iterrows():
        point = ax.scatter(hp['x'],hp['z'],
                           s=20,
                           #color=colors[index],
                           color='gainsboro',
                           edgecolors='dimgray',
                           #markeredgecolor='r',
                           label=hp['name'],
                           alpha=1,
                           zorder=3
                           )

    # plot frame
    frame_color = 'whitesmoke'
    frame_alpha = 1
    frame_width = 3


    rectangle0 = plt.Rectangle((-0.3,-0.7), 0.7, 0.58,
                          linewidth=frame_width,                          
                          ec=frame_color,
                          alpha=frame_alpha,
                          facecolor='white')
    ax.add_patch(rectangle0)

    circ = plt.Circle((0.05,-0.11), 0.11,
                      linewidth=frame_width,                          
                      ec=frame_color,
                      alpha=frame_alpha,
                      facecolor=frame_color)
    ax.add_patch(circ)
    
    # rectangle = plt.Rectangle((-0.04,-0.22), 0.22, 0.22,
    #                       linewidth=frame_width,
    #                       ec=frame_color,
    #                       alpha=frame_alpha,
    #                       facecolor=frame_color)
    

        
    # ax.add_patch(rectangle)
        
    ax.plot([-0.5, 0.63],[0,0],
      linewidth=1 ,
      alpha=frame_alpha,
      linestyle= 'solid',
      color='dimgray',
      )
    
    ax.plot([0, 0],[0,0.54],
      linewidth=1 ,
      alpha=frame_alpha,
      linestyle= 'solid',
      color='dimgray',
      )

    ax.plot([0.13, 0.13],[0,0.14],
      linewidth=1 ,
      alpha=frame_alpha,
      linestyle= 'solid',
      color='dimgray',
      )    
    
    # plot localizations
    ax.scatter(loc_data['x'], loc_data['z'],c=loc_data['time_min_offset'],
                        s=params['loc_size'].values[0],
                        marker=params['loc_marker'].values[0],
                        #color=params['loc_color'].values[0],
                        alpha=params['loc_alpha'].values[0],
                        cmap=cmap,
                        norm=norm,
                        zorder=5
                        )
    # plot uncertainties
    for idx, loc_point in loc_data.iterrows():   
        ax.plot([loc_point['x_min_CI99'],loc_point['x_max_CI99']],
                [loc_point['z'],loc_point['z']],
                linewidth=params['uncertainty_width'].values[0],
                linestyle=params['uncertainty_style'].values[0],
                #color=params['uncertainty_color'].values[0],
                color=cmap(norm(loc_point['time_min_offset'])),
                alpha=params['uncertainty_alpha'].values[0],
                )
    
        ax.plot([loc_point['x'],loc_point['x']],
                [loc_point['z_min_CI99'],loc_point['z_max_CI99']],
                linewidth=params['uncertainty_width'].values[0],
                linestyle=params['uncertainty_style'].values[0],
                #color=params['uncertainty_color'].values[0],
                color=cmap(norm(loc_point['time_min_offset'])),
                alpha=params['uncertainty_alpha'].values[0],
                )
        
    # Axes labels
    ax.set_xlabel('X (m)', labelpad=4)
    ax.set_ylabel('Z (m)', labelpad=4)
    ax.set_xlim(params['x_min'].values[0], params['x_max'].values[0])
    ax.set_ylim(params['z_min'].values[0], params['z_max'].values[0])
    #ax.grid()
    ax.set_aspect('equal', adjustable='box')
    return ax

def plot_video_frame(video_file,frame_time_sec, ax):
    frame = []
    cap = cv2.VideoCapture(video_file)
    if (cap.isOpened()== False):        
        print("Error opening video stream or file")
    else:
     	# Read until video is completed
        idx=0
        while(cap.isOpened()):            
            ret, frame = cap.read()            
            if ret == True:
                idx+=1
                fps = cap.get(cv2.CAP_PROP_FPS)  # OpenCV2 version 2 used "CV_CAP_PROP_FPS"                
                current_time = idx*(1/fps)
                if current_time >= frame_time_sec:
                    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB) 
                    frame = frame[90:900,340:1220]
                    break                
            # Break the loop
            else:
                break
        cap.release()
        #cv2.destroyAllWindows()
    return ax.imshow(frame)

def calc_data_error(tdoa_sec, m, sound_speed_mps,hydrophones_config, hydrophone_pairs):
    """ Calculates tdoa measurement errors. Eq. (9) in Mouy et al. 2018"""
    tdoa_m = predict_tdoa(m, sound_speed_mps, hydrophones_config, hydrophone_pairs)
    Q = len(m)
    M = m.size # number of dimensions of the model (here: X, Y, and Z)
    N = len(tdoa_sec) # number of measurements
    if N > M:
        error_std = np.sqrt((1/(Q*(N-M))) * (sum((tdoa_sec-tdoa_m)**2)))
    else:
        error_std = np.sqrt((sum((tdoa_sec-tdoa_m)**2)))
    return error_std


def calc_loc_errors(tdoa_errors_std, m, sound_speed_mps, hydrophones_config, hydrophone_pairs):
    """ Calculates localization errors. Eq. (8) in Mouy et al. 2018."""
    A = defineJacobian(hydrophones_config, m, sound_speed_mps, hydrophone_pairs)
    Cm = (tdoa_errors_std**2) * np.linalg.inv(np.dot(A.transpose(),A)) # Model covariance matrix for IID
    err_std = np.sqrt(np.diag(Cm))
    return pd.DataFrame({'x_std': [err_std[0]], 'y_std': [err_std[1]], 'z_std': [err_std[2]]})

## ###########################################################################

def plot_full_figure(time_sec=None):
        
    loc_file = r'C:\Users\xavier.mouy\Documents\Reports_&_Papers\Papers\10-XAVarray_2020\results\mini-array_copper\localizations_2cm_3m_v5.nc'
    loc_file_matlab = r'C:\Users\xavier.mouy\Documents\Reports_&_Papers\Papers\10-XAVarray_2020\results\mini-array_copper\localizations_matlab_with_CI.csv'
    audio_file = r'C:\Users\xavier.mouy\Documents\Reports_&_Papers\Papers\10-XAVarray_2020\data\mini_array\671404070.190801232502.wav'
    video_file = r'C:\Users\xavier.mouy\Documents\Reports_&_Papers\Papers\10-XAVarray_2020\data\large_array\2019-09-15_HornbyIsland_AMAR_07-HI\3420_FishCam01_20190920T163627.613206Z_1600x1200_awb-auto_exp-night_fr-10_q-20_sh-0_b-50_c-0_i-400_sat-0.mp4'
    hp_config_file = r'C:\Users\xavier.mouy\Documents\Reports_&_Papers\Papers\10-XAVarray_2020\data\mini_array\hydrophones_config_05-MILL_corrected.csv'
    localization_config_file =r'C:\Users\xavier.mouy\Documents\Reports_&_Papers\Papers\10-XAVarray_2020\config_files\localization_config_mini_array.yaml'    
    t1_sec = 696+15 #696
    #t1_sec = 696+14 #696
    t2_sec = 713

    
    filter_x=[-10, 10]
    filter_y=[-10, 10]
    filter_z=[-2, 10]
    filter_x_std=10
    filter_y_std=10
    filter_z_std=10
    
    params=pd.DataFrame({
        'loc_color': ['black'],
        'loc_marker': ['o'],
        'loc_alpha': [1],
        'loc_size': [5],
        'uncertainty_color': ['black'],
        'uncertainty_style': ['-'],
        'uncertainty_alpha': [1], #0.7
        'uncertainty_width': [0.2], #0.2
        'x_min':[-0.7],
        'x_max':[0.7],
        'y_min':[-0.8],
        'y_max':[0.8],
        'z_min':[-0.8],
        'z_max':[0.8],    
        })
        
    ## ###########################################################################
    localization_config = read_yaml(localization_config_file)
    hydrophones_config = pd.read_csv(hp_config_file)
    sound_speed_mps = localization_config['ENVIRONMENT']['sound_speed_mps']
    ref_channel = localization_config['TDOA']['ref_channel']
    hydrophone_pairs = defineReceiverPairs(len(hydrophones_config), ref_receiver=ref_channel)
    
    ## load localization results
    loc = Measurement()
    loc.from_netcdf(loc_file)
    loc_data = loc.data
    
    # used matlab CI
    loc_data = pd.read_csv(loc_file_matlab)
    
    # ## recalculate data errors
    # diff=[]
    # idx = 0
    # for idx in range(len(loc_data)):
    #     m = loc_data.loc[[idx],['x','y','z']]        
    #     tdoa_m = predict_tdoa(m, sound_speed_mps, hydrophones_config, hydrophone_pairs)
    #     tdoa_measured = loc_data.loc[[idx],['tdoa_sec_1','tdoa_sec_2','tdoa_sec_3']].to_numpy()    
    #     #diff_temp = (tdoa_m-tdoa_measured.T)**2
    #     if idx==0:
    #         diff = (tdoa_m-tdoa_measured.T)**2
    #     else:
    #         diff = np.vstack((diff,(tdoa_m-tdoa_measured.T)**2))
    
    # Q = len(loc_data)
    # #M = m.size # number of dimensions of the model (here: X, Y, and Z)
    # #N = len(tdoa_sec) # number of measurements    
    # #error_std = np.sqrt((1/(Q*(N-M))) * (sum((tdoa_sec-tdoa_m)**2)))    
    # tdoa_errors_std = np.sqrt( (1/Q)*(sum(diff)))
    
    # #tdoa_errors_std = calc_data_error(tdoa_sec, m, sound_speed_mps,hydrophones_config, hydrophone_pairs)
    # for idx in range(len(loc_data)):
    #     loc_errors_std = calc_loc_errors(tdoa_errors_std, loc_data.loc[[idx],['x','y','z']] , sound_speed_mps, hydrophones_config, hydrophone_pairs)
    #     print('m')
    
    # Filter
    loc_data = loc_data.dropna(subset=['x', 'y','z']) # remove NaN
    loc_data = loc_data.loc[(loc_data['x']>=min(filter_x)) & 
                            (loc_data['x']<=max(filter_x)) &
                            (loc_data['y']>=min(filter_y)) & 
                            (loc_data['y']<=max(filter_y)) &
                            (loc_data['z']>=min(filter_z)) & 
                            (loc_data['z']<=max(filter_z)) &
                            (loc_data['x_std']<= filter_x_std) & 
                            (loc_data['y_std']<= filter_y_std) &
                            (loc_data['z_std']<= filter_z_std) &
                            (loc_data['time_min_offset']>= t1_sec) &
                            (loc_data['time_max_offset']<= t2_sec)
                            #(loc_data['frequency_max']>= 60)
                            ]
    # Adjust detection times
    loc_data['time_min_offset'] = loc_data['time_min_offset'] - t1_sec
    loc_data['time_max_offset'] = loc_data['time_max_offset'] - t1_sec
    
    if time_sec!= None:
        loc_data = loc_data.loc[(loc_data['time_max_offset']<=time_sec)] 
    else:
        print('Static')
    
    # update loc object
    loc.data = loc_data
    
    # plots
    # fig, ax = plt.subplots(figsize=(6, 1))
    # fig.subplots_adjust(bottom=0.5)
    # n_colors = t2_sec-t1_sec
    # cmap = mpl.cm.get_cmap('CMRmap', n_colors*2)
    # norm = mpl.colors.Normalize(vmin=0, vmax=n_colors)
    # ax_cmap = mpl.colorbar.ColorbarBase(ax, cmap=cmap,
    #                                 norm=norm,
    #                                 orientation='horizontal')
    # ax_cmap.set_label('Time (s)')
    
   
    
    # Plot spectrogram
    fig_final, ax_spectro = plot_spectrogram(audio_file,loc,t1_sec, t2_sec, geometry=(5,1,1))    
    ax_spectro.set_title("")
    ax_spectro.get_xaxis().set_visible(False)
    n_colors = t2_sec-t1_sec
    cmap = mpl.cm.get_cmap('viridis', n_colors*4)
    norm = mpl.colors.Normalize(vmin=0, vmax=n_colors)
    divider = make_axes_locatable(ax_spectro)
    cax = divider.append_axes('bottom', 0.1, pad=0.03 )
    ax_cmap = mpl.colorbar.ColorbarBase(cax, cmap=cmap,
                                    norm=norm,
                                    orientation='horizontal')
    ax_cmap.set_label('Time (s)')
    
    if time_sec:
        SFreq_min, SFreq_max = ax_spectro.get_ylim()
        ax_spectro.plot([time_sec,time_sec],[SFreq_min,SFreq_max],'r')
    
    # plot detection points on top of spectrogram
    #gs0 = fig_final.add_gridspec(60,1)
    ax_detec = fig_final.add_subplot(20,1,1)
    det_y = np.asarray(np.ones((1,len(loc_data['time_min_offset']))))[0]
    det_x = np.asarray(loc_data['time_min_offset'])
    ax_detec.scatter(det_x,det_y, c=loc_data['time_min_offset'],cmap=cmap,norm=norm,s=12)
    ax_detec.set_xlim(ax_spectro.get_xlim())
    ax_detec.get_xaxis().set_visible(False)
    ax_detec.get_yaxis().set_visible(False)
    ax_detec.axis('off')
    
    # #pos =[left, bottom, width, height]
    # box = ax_detec.get_position()
    # box.y0 = box.y0 + 0.6
    # box.y1 = box.y1 + 0.6
    # ax_detec.set_position(box)
    
    #size = fig_final.get_size_inches()
    
    
    plt.subplots_adjust(left=0.08, bottom=0.1, right=0.95, top=0.95, wspace=0, hspace=0)
    
    
    # divider2 = make_axes_locatable(ax_spectro)
    # cax2 = divider2.append_axes('top', size=0.2, pad=10.0)
    # det_y = np.asarray(np.ones((1,len(loc_data['time_min_offset']))))[0]
    # det_x = np.asarray(loc_data['time_min_offset'])
    # cax2.plot(det_x,det_y,'.r')
    # cax2.set_xlim(ax_spectro.get_xlim())
    
    
    # ax_cmap = mpl.colorbar.ColorbarBase(cax, cmap=cmap,
    #                                     norm=norm,
    #                                     orientation='horizontal')
    
    
    
    gs = fig_final.add_gridspec(3,2)
    
    # plot localization top
    ax_toploc = fig_final.add_subplot(gs[1:,1])
    plot_top_view(hydrophones_config,loc_data,params,cmap,norm, ax_toploc)
    ax_toploc.set_anchor('E')
    
    # plot localization side
    #ax_sideloc = fig_final.add_subplot(3,3,7,sharex = ax_toploc)
    ax_sideloc = fig_final.add_subplot(gs[1:,0])
    plot_side_view(hydrophones_config,loc_data,params,cmap,norm,ax_sideloc)
    ax_sideloc.set_anchor('W')
    
    # set the spacing between subplots
    plt.subplots_adjust(wspace=0, hspace=0)
    
    
    # # plot video frame 1
    # fig_video1, ax_video1 = plt.subplots(1,1)
    # frame1_sec = 152.8 # second detection -> 16:38:59.8
    # #ax_video1 = fig_final.add_subplot(3,3,5)
    # plot_video_frame(video_file,frame1_sec, ax_video1)
    # ax_video1.get_xaxis().set_visible(False)
    # ax_video1.get_yaxis().set_visible(False)
    
    # # plot video frame 2
    # fig_video2, ax_video2 = plt.subplots(1,1)
    # frame2_sec = 160 # 4th detection -> 16:39:07
    # #ax_video2 = fig_final.add_subplot(3,3,6)
    # plot_video_frame(video_file,frame2_sec, ax_video2)
    # ax_video2.get_xaxis().set_visible(False)
    # ax_video2.get_yaxis().set_visible(False)
    
    
    fig_final.set_size_inches(9.47, 6.72)

    box = ax_spectro.get_position()
    box.y0 = box.y0 - 0.03
    box.y1 = box.y1 - 0.03
    ax_spectro.set_position(box)
    return fig_final


def main():
    
    # static
    fig = plot_full_figure()
    size = fig.get_size_inches()
    
    # # movie
    # outdir = r'C:\Users\xavier.mouy\Documents\PhD\Thesis\phd-thesis\Figures\XAV_arrays\MobileArray_Copper2\animation'
    # fps=20
    # duration_sec = 10
    
    # # create individual frames
    # times = np.arange(0,duration_sec,1/fps)
    # for idx, t in enumerate(times):
    #     plot_full_figure(time_sec=t)
    #     plt.savefig(os.path.join(outdir,str(idx)+'.jpg'))
    #     plt.close()
    
    # # stack all frames
    # img_array = []
    # for filename in range(0,idx):
    #     img = cv2.imread(os.path.join(outdir,str(filename)+'.jpg'))
    #     height, width, layers = img.shape
    #     size = (width,height)
    #     img_array.append(img)
    # # write videos
    # out = cv2.VideoWriter(os.path.join(outdir,'animation.mp4'),cv2.VideoWriter_fourcc(*'DIVX'), fps, size) 
    # for i in range(len(img_array)):
    #     out.write(img_array[i])
    # out.release()
    
if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
"""
Created on Sat Oct 19 22:22:32 2019

@author: xavier.mouy
"""

# C:\Users\xavier.mouy\Documents\Workspace\GitHub\Fish-localization\results\ArrayOptimization\run10_mean_6hp_sphere-2m

import matplotlib.pyplot as plt
import pandas as pd
from mpl_toolkits.mplot3d import Axes3D
import numpy as np
import matplotlib.animation as animation
import os
from PIL import Image
import glob
import pickle



picklefile=r'C:\Users\xavier.mouy\Documents\Workspace\GitHub\Fish-localization\results\ArrayOptimization\20191118092620_Receivers6_Bounds1m_Sources500_Radius2m\ArrayOptimizationResults_iteration-1.pickle'
outdir1 = r"C:\Users\xavier.mouy\Documents\Workspace\GitHub\Fish-localization\results\ArrayOptimization\20191118092620_Receivers6_Bounds1m_Sources500_Radius2m\animation\frames"


data = pickle.load(open(picklefile, 'rb'))
outroot = data["outroot"]
outdir = data["outdir"]
nIter = data["nIter"]
nsources = data["nsources"]
radius = data["radius"]
origin = data["origin"]
spacing = data["spacing"]
NoiseSTD = data["NoiseSTD"]
V = data["V"]
nReceivers = data["nReceivers"]
ReceiverBoundValue = data["ReceiverBoundValue"]
ReceiverBounds = data["ReceiverBounds"]
AnnealingSchedule = data["AnnealingSchedule"]
R = data["R"]
Rchanges = data["Rchanges"]
acceptRateChanges = data["acceptRateChanges"]
Cost = data["Cost"]
S = data["S"]
Uncertainties2 = data["Uncertainties2"]
                  

# set a list with acceptance rate for each iteration and add to the Cost dataFrane
acRate=[]
for t in Cost['T']:
    acRate.append(acceptRateChanges[acceptRateChanges['T']==t]['acceptRate'].iloc[0])
Cost['acceptRate']=acRate


def update(i,Cost):
    plt.close('all')
    gridsize = (3, 3)
    fig = plt.figure(figsize=(20, 10))
    fig.set_tight_layout(True)
    
    ax1 = plt.subplot2grid(gridsize, (0, 0), colspan=2, rowspan=3, projection='3d')
    ax2 = plt.subplot2grid(gridsize, (0, 2))
    ax3 = plt.subplot2grid(gridsize, (1, 2))
    ax4 = plt.subplot2grid(gridsize, (2, 2))
    
    Vmin=-1
    Vmax = 1
    #ax1.scatter(Rchanges[:,0,i], Rchanges[:,1,i], Rchanges[:,2,i], c=['r','y','k','b','g','c'],s=40)
    ax1.scatter(Rchanges[0,0,i], Rchanges[0,1,i], Rchanges[0,2,i], c=['r'],s=60,label='Hydrophone 1')
    ax1.scatter(Rchanges[1,0,i], Rchanges[1,1,i], Rchanges[1,2,i], c=['y'],s=60,label='Hydrophone 2')
    ax1.scatter(Rchanges[2,0,i], Rchanges[2,1,i], Rchanges[2,2,i], c=['k'],s=60,label='Hydrophone 3')
    ax1.scatter(Rchanges[3,0,i], Rchanges[3,1,i], Rchanges[3,2,i], c=['b'],s=60,label='Hydrophone 4')
    ax1.scatter(Rchanges[4,0,i], Rchanges[4,1,i], Rchanges[4,2,i], c=['g'],s=60,label='Hydrophone 5')
    ax1.scatter(Rchanges[5,0,i], Rchanges[5,1,i], Rchanges[5,2,i], c=['c'],s=60,label='Hydrophone 6')            
    ax1.set_xlim([Vmin,Vmax])
    ax1.set_ylim([Vmin,Vmax])
    ax1.set_zlim([Vmin,Vmax])
    ax1.set_xlabel('X (m)',fontsize=14, fontweight='bold')
    ax1.set_ylabel('Y (m)',fontsize=14, fontweight='bold')
    ax1.set_zlabel('Z (m)',fontsize=14, fontweight='bold')
    ax1.legend(loc='upper left',bbox_to_anchor= (0.0, 1.01),frameon=False,fontsize=12)
    ax1.view_init(30, 80)
    
    ax2.semilogy(Cost['T'][0:i])
    ax2.set_ylabel('Temperature',fontsize=14, fontweight='bold')
    ax2.set_xlim([0,13000])
    ax2.set_ylim([2.464046623054599e-05,200])
    
    ax3.plot(Cost['acceptRate'][0:i])
    ax3.set_ylabel('Acceptance rate',fontsize=14, fontweight='bold')
    ax3.set_xlim([0,13000])
    ax3.set_ylim([0,1])
    
    ax4.plot(Cost['cost'][0:i])
    ax4.set_xlabel('Iteration',fontsize=14, fontweight='bold')
    ax4.set_ylabel('Cost',fontsize=14, fontweight='bold')
    ax4.set_xlim([0,13000])
    ax4.set_ylim([0,50])

    filename = str(i) + '.png'
    fname = os.path.join(outdir1, filename)
    plt.savefig(fname, dpi=None, facecolor='w', edgecolor='w',
            orientation='portrait', papertype=None, format=None,
            transparent=False, bbox_inches=None, pad_inches=0.1,
            frameon=None, metadata=None)


for i in range(0,12588,50):
    update(i,Cost)

#for i in range(12500,12588,1):
#    update(i,Cost)

# Open all the frames
files = glob.glob(outdir1 + '/*.png')
files = os.listdir(outdir1)
filessorted = sorted(files,key=lambda x: int(os.path.splitext(x)[0]))

images = []
for n in filessorted:
    frame = Image.open(os.path.join(outdir1,n))
    images.append(frame)

# Save the frames as an animated GIF
images[0].save(os.path.join(outdir1,'animation.gif'),
               save_all=True,
               append_images=images[1:],
               duration=200,
               loop=1)
    


# -*- coding: utf-8 -*-
"""
Created on Tue Aug 29 09:35:38 2017

@author: xavier
"""
import numpy as np
import pandas as pd
import pickle
import os
import mpl_toolkits.mplot3d
import matplotlib.pyplot as plt
import datetime

import localizationlib as loclib

#def defineSphereSurfaceGrid(npoints, radius, origin=[0, 0, 0]):
#    # Using the golden spiral method
#    # ------------------
#    # inputs:
#    #   npoints =>  nb of points on the sphere - integer
#    #   radius  => radius of the sphere - float
#    #   origin  => origin of teh sphere in catesian coordinates- 3 element list
#    # ------------------
#    # sampling in spherical coordinates
#    indices = np.arange(0, npoints, dtype=float) + 0.5
#    phi = np.arccos(1 - 2*indices/npoints)
#    theta = np.pi * (1 + 5**0.5) * indices
#    # convert to cartesian coordinates
#    Sx, Sy, Sz = radius*np.cos(theta) * np.sin(phi), radius*np.sin(theta) * np.sin(phi), radius*np.cos(phi)
#    # Adjust origin
#    Sx = Sx + origin[0]
#    Sy = Sy + origin[1]
#    Sz = Sz + origin[2]
#    # package in a datafrane
#    S = pd.DataFrame({'x': Sx, 'y': Sy, 'z': Sz})
#    return S
#
#
#def defineSphereVolumeGrid(spacing, radius, origin=[0, 0, 0]):
#    # ------------------
#    # inputs:
#    #   spacing =>  distance in meters separatying each receiver - float
#    #   radius  => radius of the sphere - float
#    #   origin  => origin of the sphere in catesian coordinates- 3 element list
#    # ------------------
#    # Cube of points (Cartesian coordinates)
#    vec = np.arange(-radius, radius, spacing)
#    X, Y, Z = np.meshgrid(vec, vec, vec, indexing='ij')
#    Sx = np.reshape(X, X.shape[0]*X.shape[1]*X.shape[2])
#    Sy = np.reshape(Y, Y.shape[0]*Y.shape[1]*Y.shape[2])
#    Sz = np.reshape(Z, Z.shape[0]*Z.shape[1]*Z.shape[2])
#    # Convert to spherical coordinates and remove points with r < radius
#    Sr = np.sqrt(Sx**2 + Sy**2 + Sz**2)
#    Sr_sphere = Sr <= radius
#    Sx_sphere = Sx[Sr_sphere]
#    Sy_sphere = Sy[Sr_sphere]
#    Sz_sphere = Sz[Sr_sphere]
#    # Adjust origin
#    Sx_sphere = Sx_sphere + origin[0]
#    Sy_sphere = Sy_sphere + origin[1]
#    Sz_sphere = Sz_sphere + origin[2]
#    # package in a datafrane
#    S = pd.DataFrame({'x': Sx_sphere, 'y': Sy_sphere, 'z': Sz_sphere})
#    return S
#
#
#def defineReceiverPairs (nReceivers):
#    refReceiver = 1
#    Rpairs = []
#    for i in range(nReceivers):
#        if i != refReceiver:
#            pair = [refReceiver, i]
#            Rpairs.append(pair)
#    return Rpairs
#
#
#def defineJacobian(R, S, V, Rpairs):
#    N = R.shape[0] - 1          # nb of measurements (TDOAs)
#    M = 3                       # number of model parameters (unknowns)
#    nsources = S.shape[0]       # number of sources
#    J = [None] * nsources       # initiaization
#    # for each source location
#    for idx in range(nsources):
#        s = S.iloc[idx]
#        j = np.full([N, M], np.nan)  # initialization of Jacobian for that source location
#        for i in range(N):
#            p1 = Rpairs[i][0]        # receiver #1 ID
#            p2 = Rpairs[i][1]        # receiver #2 ID
#            for kk in range(M):
#                Term1 = (1/V)*0.5*((((s.x-R.x[p1])**2)+((s.y-R.y[p1])**2)+((s.z-R.z[p1])**2))**(-0.5))*2*(s.iloc[kk]-R.iloc[p1][kk])
#                Term2 = (1/V)*0.5*((((s.x-R.x[p2])**2)+((s.y-R.y[p2])**2)+((s.z-R.z[p2])**2))**(-0.5))*2*(s.iloc[kk]-R.iloc[p2][kk])
#                j[i][kk] = Term2 - Term1
#        J[idx] = j  # stacks jacobians for each source
#    return J
#
#
#def getUncertainties(J, NoiseVariance):
#    nsources = len(J)
#    errLoc_X = [None] * nsources
#    errLoc_Y = [None] * nsources
#    errLoc_Z = [None] * nsources
#    errLoc_RMS = [None] * nsources
#    for i in range(nsources):
#        Cm = NoiseVariance * np.linalg.pinv(np.dot(np.transpose(J[i]), J[i]))  # covariance matrix of the model
#        errLoc_X[i], errLoc_Y[i], errLoc_Z[i] = np.sqrt(np.diag(Cm))  # uncertainty (std) along each axis
#        errLoc_RMS[i] = np.sqrt(errLoc_X[i]**2 + errLoc_Y[i]**2 + errLoc_Z[i]**2)  # overall uncertainty (RMS)
#    Uncertainty = pd.DataFrame({'x': errLoc_X, 'y': errLoc_Y, 'z': errLoc_Z, 'rms': errLoc_RMS})
#    return Uncertainty
#
#
#def plotArrayUncertainties(R, S, Uncertainties):
#    fig1 = plt.figure()
#    ax1 = fig1.add_subplot(111, projection='3d')
#    # Sources
#    points = ax1.scatter(S['x'], S['y'], S['z'], s=10, c=Uncertainties.rms.values, cmap='spectral')
#    cbar = plt.colorbar(points)
#    cbar.ax.set_ylabel('RMS uncertainty (m)')
#    # Receivers
#    ax1.scatter(R['x'], R['y'], R['z'], s=30, c='black')
#    # Axes labels
#    ax1.set_xlabel('X (m)', labelpad=10)
#    ax1.set_ylabel('Y (m)', labelpad=10)
#    ax1.set_zlabel('Z (m)', labelpad=10)
#    plt.show()
#
#
#def getCost(R, S, Rpairs, V, NoiseVariance):
#    # Get list of Jacobian matrice for each source
#    J = defineJacobian(R, S, V, Rpairs)
#    # Calculates localization uncertainty for each source
#    Uncertainties = getUncertainties(J, NoiseVariance)
#    # Get max uncertainty
#    #E = max(Uncertainties.rms)
#    E = np.mean(Uncertainties.rms)
#    #E = np.median(Uncertainties.rms)
#    return E
#
#
#def getReceiverBoundsWidth(ReceiverBounds):
#    ReceiverBoundsWidth = ReceiverBounds.applymap(lambda x: max(x)-min(x))
#    return ReceiverBoundsWidth
#
#
#def initializeReceivers(nReceivers, ReceiverBounds):
#    ReceiverBoundsWidth = getReceiverBoundsWidth(ReceiverBounds)
#    R1 = [None] * ReceiverBounds.shape[1]
#    R1[0] = [None] * nReceivers  # x
#    R1[1] = [None] * nReceivers  # y
#    R1[2] = [None] * nReceivers  # z
#    for r in range(nReceivers):  # random location for each receiver and axis
#        for dim in range(ReceiverBounds.shape[1]):
#            R1[dim][r] = np.random.rand(1)[0]*ReceiverBoundsWidth.iloc[r][dim] + min(ReceiverBounds.iloc[r][dim])
#    Receivers = pd.DataFrame({'x': R1[0], 'y': R1[1], 'z': R1[2]})
#    return Receivers
#
#
#def getParamsLinearMapping(R):
#    Rindices = [None] * R.shape[0] * R.shape[1]
#    midx = 0
#    for ridx in range(R.shape[0]):
#        for dimidx in range(R.shape[1]):
#            Rindices[midx] = [ridx, dimidx]
#            midx += 1
#    return Rindices
#
#
#def perturbReceivers(R, PerturbParamIdx, MappedParamsIdx, ReceiverBounds, ReceiverBoundsWidth, PerturbSTD, T0, T):
#    # goes back to first parameters if reached the end of the list of parameters
#    if PerturbParamIdx > len(MappedParamsIdx)-1:
#        PerturbParamIdx = 0
#    # Identifies from MappedParamsIdx which Receiver and Dimension to perturb
#    rid = MappedParamsIdx[PerturbParamIdx][0]    # Receiver ID
#    dimid = MappedParamsIdx[PerturbParamIdx][1]  # Dimension ID
#    # Add perturbation to parameter
#    perturb = ((PerturbSTD*ReceiverBoundsWidth.iloc[rid][dimid])*np.random.normal(loc=0))  # Gaussian distributed perturbation
#    # nu = np.random.random()
#    # gamma = PerturbSTD*(T/T0)*np.tan(np.pi*(nu-0.5))
#    # bandwidth = ReceiverBoundsWidth.iloc[rid][dimid]
#    # perturb = gamma*bandwidth
#    newparam = R.iloc[rid][dimid] + perturb
#    # print(perturb)
#    # Checks that perturbed parameter lies within the bounds
#    isinbound = (newparam >= min(ReceiverBounds.iloc[rid][dimid])) & (newparam <= max(ReceiverBounds.iloc[rid][dimid]))
#    # updates receiver parameter (only if new paramater fall within parameter bounds)
#    R_prime = pd.DataFrame.copy(R)
#    if isinbound == True:
#        R_prime.iloc[rid][dimid] = newparam        
#    return R_prime, isinbound, PerturbParamIdx
#
#
#def optimizeArray(ReceiverBounds, nReceivers, AnnealingSchedule, S, Rpairs, V, NoiseVariance):
#    # start clock
#    start = time.time()
#    
#    # Defines width of parameters bounds
#    ReceiverBoundsWidth = getReceiverBoundsWidth(ReceiverBounds)
#    
#    # initialization of variables    
#    Cost = pd.DataFrame({'T': [], 'cost': []})
#    acceptRateChanges = pd.DataFrame({'T': [], 'acceptRate': []})
#    acceptRate = 1
#    PerturbParamIdx = -1
#    Tidx = 0  # temperature step index
#    LoopStopFlag = 0
#    Rchanges = []
#    while LoopStopFlag == 0:  # Temperature loop. Keeps iterating until acceptance rate is too low
#        # for nnn in range(1):
#        # First iteration
#        if (Tidx == 0):
#            R = initializeReceivers(nReceivers, ReceiverBounds)  # random initialization of receivers locations (whithin the bounds)
#            E_m = getCost(R, S, Rpairs, V, NoiseVariance)        # Calculates max RMS uncertainty
#            T = AnnealingSchedule['Start']                       # initial temperature
#            tmp1 = pd.DataFrame({'T': [T], 'cost': [E_m]})
#            Cost = pd.DataFrame.append(Cost, tmp1, ignore_index=True)
#            Rchanges = R.as_matrix()                                         # Keeps track of model paraneters at each iteration
#            MappedParamsIdx = getParamsLinearMapping(R)          # linear list of each elements to optimize (for the perturnation phase)
#            #plotArrayUncertainties(R, S, Uncertainties)        
#
#        # Checks that starting temperature is not set to low
#        if (Tidx == 1):
#            if (acceptRate < AnnealingSchedule['StartAcceptanceRate']):
#                raise ValueError(['The acceptance rate during the melting phase is too low.Please adjust starting temperature (' + str(acceptRate) + ')'])    
#            else:
#                print('Melting temperature valid (' + str(acceptRate) + ').')
#
#        # Perturb paraneters
#        nAccepted = 0  # keeps track of accepted perturbations
#        for j in range(AnnealingSchedule['nPerturb']):  # perturbations of parameters for temperature T
#            # Perturb one receiver parameter
#            PerturbParamIdx += 1  # increment to next parameter in line
#            R_prime, isinbound, PerturbParamIdx = perturbReceivers(R, PerturbParamIdx, MappedParamsIdx, ReceiverBounds, ReceiverBoundsWidth, AnnealingSchedule['PerturbSTD'], AnnealingSchedule['Start'], T)
#            # Acceptance tests
#            acceptedFlag = []
#            if isinbound:
#                E_mprime = getCost(R_prime, S, Rpairs, V, NoiseVariance)       # Calculates max RMS uncertainty
#                deltaE = E_mprime-E_m
#                if (deltaE <= 0):   # accept change
#                    acceptedFlag = True
#                elif (deltaE > 0):
#                    psy = np.random.uniform()
#                    P = np.exp(-deltaE/T)
#                    if (psy <= P):  # accept change
#                        acceptedFlag = True
#                    else:           # reject change
#                        acceptedFlag = False
#            else:
#                acceptedFlag = False
#            # Updates parmaters for accepted changes
#            if acceptedFlag:  # accepted change
#                nAccepted += 1
#                # update paramters
#                R = pd.DataFrame.copy(R_prime)
#                E_m = E_mprime
#                del R_prime, E_mprime   # delete variables
#                # saves cost and model parameters
#                tmp1 = pd.DataFrame({'T': [T], 'cost': [E_m]})
#                Cost = pd.DataFrame.append(Cost, tmp1, ignore_index=True)
#                Rchanges = np.dstack((Rchanges, R.as_matrix()))
#            elif acceptedFlag is not False:  # sanity check
#                raise ValueError(['Perturbation did go throught the acceptance test. Check the code for errors!'])
#    
#        # Calculates acceptance rate for that temperature value
#        acceptRate = nAccepted / AnnealingSchedule['nPerturb']
#        tmp2 = pd.DataFrame({'T': [T], 'acceptRate': [acceptRate]})
#        acceptRateChanges = acceptRateChanges.append(tmp2, ignore_index=True)
#        print('Temperature: %.3f - Acceptance rate: %.2f - Cost: %.2f' % (T, acceptRate, E_m))
#    
#        # Stopping conditions
#        if (acceptRate < AnnealingSchedule['StopAcceptanceRate']):
#            LoopStopFlag = 1
#            print('Optimization complete (acceptance rate threshold reached)')
#        if (E_m <= AnnealingSchedule['StopCost']):
#            LoopStopFlag = 1
#            print('Optimization complete (cost objective reached)')
#    
#        # Update temperature
#        T = T * AnnealingSchedule['ReducFactor']  # decrease temperature
#        Tidx += 1  # next temperature step
#
#    end = time.time()
#    elapsedTime = end - start
#    return R, Rchanges, acceptRateChanges, Cost, elapsedTime
#
#def plotOptimizationResults(outdir, nReceivers, Rchanges, Cost, acceptRateChanges, R):
#    
#    # plot Parameters evolution with Temperature
#    f1 = plt.figure(1)
#    for Ridx in range(nReceivers):
#        plt.subplot(nReceivers,1, Ridx+1)
#        plt.plot(Rchanges[Ridx,0,:], label='X(m)', color='black')
#        plt.plot(Rchanges[Ridx,1,:], label='Y(m)', color='red')
#        plt.plot(Rchanges[Ridx,2,:], label='Z(m)', color='green')
#        plt.grid('on')
#        plt.ylabel('H' + str(Ridx+1) )
#        if Ridx == nReceivers -1:
#            plt.xlabel('Temperature step')
#        #if Ridx == 0:
#        #    plt.legend(loc="best", labels=['X(m)','Y(m)','Z(m)'], bbox_to_anchor=(0.5,-0.1))
#    f1.savefig(os.path.join(outdir, 'ReceiversPositionVsTemperature' + '_iteration-' + str(i+1) + '.png'), bbox_inches='tight')
#    
#    # plot cost evolution with Temperature
#    f2 = plt.figure(2)
#    plt.plot(Cost['cost'], color = 'black')
#    plt.grid('on')
#    plt.xlabel('Temperature step')
#    plt.ylabel('Cost')
#    f2.savefig(os.path.join(outdir, 'CostVsTemperature' + '_iteration-' + str(i+1) + '.png'), bbox_inches='tight')
#
#    
#    # plot acceptance rate with Temperature
#    f3 = plt.figure(3)
#    plt.semilogx(acceptRateChanges['T'],acceptRateChanges['acceptRate'], color = 'black')
#    plt.grid('on')
#    plt.xlabel('Temperature')
#    plt.ylabel('Acceptance rate')
#    plt.semilogx
#    f3.savefig(os.path.join(outdir, 'AcceptanceRateVsTemperature' + '_iteration-' + str(i+1) + '.png'), bbox_inches='tight')
#
#    
#    # plot Final receivers positions
#    f4 = plt.figure(4)
#    ax1 = f4.add_subplot(111, projection='3d')
#    # Receivers
#    ax1.scatter(R['x'], R['y'], R['z'], s=30, c='black')
#    # Axes labels
#    ax1.set_xlabel('X (m)', labelpad=10)
#    ax1.set_ylabel('Y (m)', labelpad=10)
#    ax1.set_zlabel('Z (m)', labelpad=10)
#    plt.show()
#    f4.savefig(os.path.join(outdir, 'FinalReceiversPosition' + '_iteration-' + str(i+1) + '.png'), bbox_inches='tight')

# ------------------------------------------------------
# input parameters --------------------------------------

# Results directory
outroot = r'C:\Users\xavier.mouy\Documents\Workspace\GitHub\Fish-localization\results\ArrayOptimization'

# Number of iterations (nb of time the process is repeated to ensure stability of the solution)
nIter = 1

## Spherical grid (sources location)
nsources = 500 #300
radius = 2
origin = [0, 0, 0]
spacing = 0.5

## cubic grid
#radius = 2
#origin = [0, 0, 0]
#spacing = 0.1 # 0.5

# Measurements errors (TDOA)
NoiseSTD = 5.0714e-04  # standard deviation of TDOAs

# Sound speed (m/s)
V = 1488

# number of receivers (for optimization)
nReceivers = 6

# Optimization
ReceiverBoundValue = 1
tmp = [[-ReceiverBoundValue, ReceiverBoundValue]] * nReceivers  # boundaries of receiver location (m)
ReceiverBounds = pd.DataFrame({'x': tmp, 'y': tmp, 'z': tmp})
#
#tmpx = [[-0.6, 0.6]] * nReceivers  # boundaries of receiver location (m)
#tmpy = [[-0.6, 0.6]] * nReceivers  # boundaries of receiver location (m)
#tmpz = [[0.6, 0.6]] * nReceivers  # boundaries of receiver location (m)
#ReceiverBounds = pd.DataFrame({'x': tmpx, 'y': tmpy, 'z': tmpz})

AnnealingSchedule = ({          # defines annealing schedule
    'Start': 200, #200
    'ReducFactor': 0.9, # before was 0.9
    'nPerturb': 100, # 30
    'PerturbSTD': 1/8,  #  1/8
    'StartAcceptanceRate': 0.8, # 0.8
    'StopAcceptanceRate': 0.001,  # 0.001
    'StopCost': 0
     })
# ------------------------------------------------------
# ------------------------------------------------------

# prelim
NoiseVariance = NoiseSTD**2

# Set virtual sources location (spherical grid)
S = loclib.defineSphereSurfaceGrid(nsources, radius, origin)
#S = loclib.defineCubeVolumeGrid(spacing, radius, origin)
nsources = S.shape[0]

# creates output folder
StartTimestamp_obj = datetime.datetime.now()
StartTimestamp_str = StartTimestamp_obj.strftime("%Y%m%d%H%M%S")
outdir = os.path.join(outroot,StartTimestamp_str + '_' + 'Receivers' + str(nReceivers) + '_' + 'Bounds' + str(ReceiverBoundValue) + 'm_' + 'Sources' + str(nsources)  + '_' + 'Radius' + str(radius) + 'm' )
os.mkdir(outdir)

# Define receiver pairs for TDOAs
Rpairs = loclib.defineReceiverPairs (nReceivers)

# Repeats optimization nIter times to ensure stability
for i in range(nIter):    
       
    # Closes all open figures
    plt.close("all")
    # Optimize array configuration 
    R, Rchanges, acceptRateChanges, Cost, processingTime = loclib.optimizeArray(ReceiverBounds, nReceivers, AnnealingSchedule, S, Rpairs, V, NoiseVariance)       
    # Get list of Jacobian matrice for each source
    J2 = loclib.defineJacobian(R, S, V, Rpairs) 
    # Calculates localization uncertainty for each source
    Uncertainties2 = loclib.getUncertainties(J2, NoiseVariance)   
    # Plots unceratinties of optimized array
    loclib.plotArrayUncertainties(R, S, Uncertainties2)
    plt.savefig(os.path.join(outdir, 'UncertaintiesPlot' + '_iteration-' + str(i+1) + '.png'), bbox_inches='tight')
    # Plots Optimization results
    loclib.plotOptimizationResults(outdir, nReceivers, Rchanges, Cost, acceptRateChanges, R, i)

    # Save paraneters and results to pickle file
    outfilename = os
    data = {"outroot": outroot , 
             "outdir": outdir , 
             "nIter": nIter , 
             "nsources": nsources , 
             "radius": radius, 
             "origin": origin, 
             "spacing": spacing, 
             "NoiseSTD": NoiseSTD, 
             "V": V, 
             "nReceivers": nReceivers, 
             "ReceiverBoundValue": ReceiverBoundValue,
             "ReceiverBounds": ReceiverBounds, 
             "AnnealingSchedule": AnnealingSchedule, 
             "R": R, 
             "Rchanges": Rchanges, 
             "acceptRateChanges": acceptRateChanges, 
             "Cost": Cost, 
             "S": S, 
             "Uncertainties2": Uncertainties2,
             "processingTime": processingTime
             }
    pickle.dump(data, open(os.path.join(outdir, 'ArrayOptimizationResults' + '_iteration-' + str(i+1) + '.pickle'), "wb"))
	



## ------------------------------------------------------
## Original array configuration -------------------------
#
## Original receivers location
#Rx = [1.185, 0, -1.185, 0]          # X values
#Ry = [0, 1.190, 0, -1.190]  # Y values
#Rz = [0, 0, 0, 0]               # Z values
#
#
## set receivers location in pandas dataframe
#R0 = pd.DataFrame({'x': Rx, 'y': Ry, 'z': Rz})
#
## Set sources location (spherical grid)
#S = defineSphereSurfaceGrid(nsources, radius, origin)
##S2 = pd.DataFrame.append(S,S1, ignore_index = True)
##S = defineSphereVolumeGrid(spacing, radius, origin)
#
## Get list of Jacobian matrice for each source
#J = defineJacobian(R0, S, V, Rpairs)
## Calculates localization uncertainty for each source
#Uncertainties = getUncertainties(J, NoiseVariance)
## plot Sources, Receivers and uncertainties
#plotArrayUncertainties(R0, S, Uncertainties)

## ------------------------------------------------------
